from typing import Tuple, Set, Dict, Optional, cast

from pyteal.ast import ScratchSlot, SubroutineDefinition
from pyteal.ir import TealBlock, Op
from pyteal.errors import TealInternalError
from pyteal.config import NUM_SLOTS


def collect_unoptimized_slots(
    subroutineBlocks: Dict[Optional[SubroutineDefinition], TealBlock]
) -> Set[ScratchSlot]:
    """Find and return all referenced ScratchSlots that need to be skipped
    during optimization.

    Args:
        subroutineBlocks: A mapping from subroutine to the subroutine's control flow graph.
        The key None is taken to mean the main program routine.

    Returns:
        A set which contains the slots used by DynamicScratchVars, all the reserved slots,
            and all global slots.
    """

    unoptimized_slots: Set[ScratchSlot] = set()

    def collectSlotsFromBlock(block: TealBlock):
        for op in block.ops:
            for slot in op.getSlots():
                # dynamic slot or reserved slot
                if op.op == Op.int or slot.isReservedSlot:
                    unoptimized_slots.add(slot)

    for _, start in subroutineBlocks.items():
        for block in TealBlock.Iterate(start):
            collectSlotsFromBlock(block)

    global_slots, _ = collectScratchSlots(subroutineBlocks)
    unoptimized_slots.update(global_slots)
    return unoptimized_slots


def collectScratchSlots(
    subroutineBlocks: Dict[Optional[SubroutineDefinition], TealBlock]
) -> Tuple[Set[ScratchSlot], Dict[Optional[SubroutineDefinition], Set[ScratchSlot]]]:
    """Find and return all referenced ScratchSlots for each subroutine.

    Args:
        subroutineBlocks: A mapping from subroutine to the subroutine's control flow graph.
        The key None is taken to mean the main program routine.

    Returns:
        A tuple of a set containing all global slots and a dictionary whose keys are the
            same as subroutineBlocks, and whose values are the local slots of that
            subroutine.
    """

    subroutineSlots: Dict[Optional[SubroutineDefinition], Set[ScratchSlot]] = dict()

    def collectSlotsFromBlock(block: TealBlock, slots: Set[ScratchSlot]):
        for op in block.ops:
            for slot in op.getSlots():
                slots.add(slot)

    for subroutine, start in subroutineBlocks.items():
        slots: Set[ScratchSlot] = set()
        for block in TealBlock.Iterate(start):
            collectSlotsFromBlock(block, slots)

        subroutineSlots[subroutine] = slots

    # all scratch slots referenced by more than 1 subroutine
    global_slots: Set[ScratchSlot] = set()

    # all scratch slots referenced by only 1 subroutine
    local_slots: Dict[Optional[SubroutineDefinition], Set[ScratchSlot]] = dict()

    for subroutine, slots in subroutineSlots.items():
        allOtherSlots: Set[ScratchSlot] = set()

        for otherSubroutine, otherSubroutineSlots in subroutineSlots.items():
            if subroutine is not otherSubroutine:
                allOtherSlots |= otherSubroutineSlots

        global_slots |= slots & allOtherSlots
        local_slots[subroutine] = slots - global_slots

    return global_slots, local_slots


def assignScratchSlotsToSubroutines(
    subroutineBlocks: Dict[Optional[SubroutineDefinition], TealBlock],
) -> Dict[Optional[SubroutineDefinition], Set[int]]:
    """Assign scratch slot values for an entire program.

    Args:
        subroutineBlocks: A mapping from subroutine to the control flow graph of the subroutine's
            blocks. The key None is taken to mean the main program routine. The values of this
            map will be modified in order to assign specific slot values to all referenced scratch
            slots.

    Raises:
        TealInternalError: if the scratch slots referenced by the program do not fit into 256 slots,
            or if multiple ScratchSlots request the same slot ID.

    Returns:
        A dictionary whose keys are the same as subroutineBlocks, and whose values are sets of
        integers representing the assigned IDs of slots which appear only in that subroutine
        (subroutine local slots).
    """
    global_slots, local_slots = collectScratchSlots(subroutineBlocks)
    # all scratch slots referenced by the program
    allSlots: Set[ScratchSlot] = global_slots | cast(Set[ScratchSlot], set()).union(
        *local_slots.values()
    )

    slotAssignments: Dict[ScratchSlot, int] = dict()
    slotIds: Set[int] = set()

    for slot in allSlots:
        if not slot.isReservedSlot:
            continue

        # If there are two unique slots with same IDs, raise an error
        if slot.id in slotIds:
            raise TealInternalError(
                "Slot ID {} has been assigned multiple times".format(slot.id)
            )
        slotIds.add(slot.id)

    if len(allSlots) > NUM_SLOTS:
        # TODO: identify which slots can be reused
        # subroutines which never invoke each other can use the same slot ID for local slots
        raise TealInternalError(
            "Too many slots in use: {}, maximum is {}".format(len(allSlots), NUM_SLOTS)
        )

    # verify that all local slots are assigned to before being loaded.
    # TODO: for simplicity, the current implementation does not perform this check with global slots
    # as well, but that would be a good improvement
    for subroutine, start in subroutineBlocks.items():
        errors = start.validateSlots(slotsInUse=global_slots)
        if len(errors) > 0:
            msg = "Encountered {} error{} when assigning slots to subroutine".format(
                len(errors), "s" if len(errors) != 1 else ""
            )
            raise TealInternalError(msg) from errors[0]

    nextSlotIndex = 0
    for slot in sorted(allSlots, key=lambda slot: slot.id):
        # Find next vacant slot that compiler can assign to
        while nextSlotIndex in slotIds:
            nextSlotIndex += 1

        if slot.isReservedSlot:
            # Slot ids under 256 are manually reserved slots
            slotAssignments[slot] = slot.id
        else:
            slotAssignments[slot] = nextSlotIndex
            slotIds.add(nextSlotIndex)

    for start in subroutineBlocks.values():
        for block in TealBlock.Iterate(start):
            for op in block.ops:
                for slot in op.getSlots():
                    op.assignSlot(slot, slotAssignments[slot])

    assignedLocalSlots: Dict[Optional[SubroutineDefinition], Set[int]] = dict()
    for subroutine, slots in local_slots.items():
        assignedLocalSlots[subroutine] = set(slotAssignments[slot] for slot in slots)

    return assignedLocalSlots



import pytest

import pyteal as pt

options = pt.CompileOptions()


def test_cond_one_pred():
    expr = pt.Cond([pt.Int(1), pt.Int(2)])
    assert expr.type_of() == pt.TealType.uint64

    cond1, _ = pt.Int(1).__teal__(options)
    pred1, _ = pt.Int(2).__teal__(options)
    cond1Branch = pt.TealConditionalBlock([])
    cond1.setNextBlock(cond1Branch)
    cond1Branch.setTrueBlock(pred1)
    cond1Branch.setFalseBlock(pt.Err().__teal__(options)[0])
    pred1.setNextBlock(pt.TealSimpleBlock([]))
    expected = cond1

    actual, _ = expr.__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_cond_two_pred():
    expr = pt.Cond([pt.Int(1), pt.Bytes("one")], [pt.Int(0), pt.Bytes("zero")])
    assert expr.type_of() == pt.TealType.bytes

    cond1, _ = pt.Int(1).__teal__(options)
    pred1, _ = pt.Bytes("one").__teal__(options)
    cond1Branch = pt.TealConditionalBlock([])
    cond2, _ = pt.Int(0).__teal__(options)
    pred2, _ = pt.Bytes("zero").__teal__(options)
    cond2Branch = pt.TealConditionalBlock([])
    end = pt.TealSimpleBlock([])

    cond1.setNextBlock(cond1Branch)
    cond1Branch.setTrueBlock(pred1)
    cond1Branch.setFalseBlock(cond2)
    pred1.setNextBlock(end)

    cond2.setNextBlock(cond2Branch)
    cond2Branch.setTrueBlock(pred2)
    cond2Branch.setFalseBlock(pt.Err().__teal__(options)[0])
    pred2.setNextBlock(end)

    expected = cond1

    actual, _ = expr.__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_cond_three_pred():
    expr = pt.Cond(
        [pt.Int(1), pt.Int(2)], [pt.Int(3), pt.Int(4)], [pt.Int(5), pt.Int(6)]
    )
    assert expr.type_of() == pt.TealType.uint64

    cond1, _ = pt.Int(1).__teal__(options)
    pred1, _ = pt.Int(2).__teal__(options)
    cond1Branch = pt.TealConditionalBlock([])
    cond2, _ = pt.Int(3).__teal__(options)
    pred2, _ = pt.Int(4).__teal__(options)
    cond2Branch = pt.TealConditionalBlock([])
    cond3, _ = pt.Int(5).__teal__(options)
    pred3, _ = pt.Int(6).__teal__(options)
    cond3Branch = pt.TealConditionalBlock([])
    end = pt.TealSimpleBlock([])

    cond1.setNextBlock(cond1Branch)
    cond1Branch.setTrueBlock(pred1)
    cond1Branch.setFalseBlock(cond2)
    pred1.setNextBlock(end)

    cond2.setNextBlock(cond2Branch)
    cond2Branch.setTrueBlock(pred2)
    cond2Branch.setFalseBlock(cond3)
    pred2.setNextBlock(end)

    cond3.setNextBlock(cond3Branch)
    cond3Branch.setTrueBlock(pred3)
    cond3Branch.setFalseBlock(pt.Err().__teal__(options)[0])
    pred3.setNextBlock(end)

    expected = cond1

    actual, _ = expr.__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected


def test_cond_has_return():
    exprWithReturn = pt.Cond(
        [pt.Int(1), pt.Return(pt.Int(1))], [pt.Int(0), pt.Return(pt.Int(0))]
    )
    assert exprWithReturn.has_return()

    exprWithoutReturn = pt.Cond(
        [pt.Int(1), pt.Bytes("one")], [pt.Int(0), pt.Bytes("zero")]
    )
    assert not exprWithoutReturn.has_return()

    exprSemiReturn = pt.Cond(
        [pt.Int(1), pt.Return(pt.Int(1))],
        [pt.Int(0), pt.App.globalPut(pt.Bytes("key"), pt.Bytes("value"))],
    )
    assert not exprSemiReturn.has_return()


def test_cond_invalid():
    with pytest.raises(pt.TealInputError):
        pt.Cond()

    with pytest.raises(pt.TealInputError):
        pt.Cond([])

    with pytest.raises(pt.TealInputError):
        pt.Cond([pt.Int(1)], [pt.Int(2), pt.Pop(pt.Txn.receiver())])

    with pytest.raises(pt.TealTypeError):
        pt.Cond([pt.Int(1), pt.Int(2)], [pt.Int(2), pt.Txn.receiver()])

    with pytest.raises(pt.TealTypeError):
        pt.Cond([pt.Arg(0), pt.Int(2)])

    with pytest.raises(pt.TealTypeError):
        pt.Cond([pt.Int(1), pt.Int(2)], [pt.Int(2), pt.Pop(pt.Int(2))])

    with pytest.raises(pt.TealTypeError):
        pt.Cond([pt.Int(1), pt.Pop(pt.Int(1))], [pt.Int(2), pt.Int(2)])


def test_cond_two_pred_multi():
    args = [
        pt.Int(1),
        [pt.Pop(pt.Int(1)), pt.Bytes("one")],
        pt.Int(0),
        [pt.Pop(pt.Int(2)), pt.Bytes("zero")],
    ]
    expr = pt.Cond(
        [args[0]] + args[1],
        [args[2]] + args[3],
    )
    assert expr.type_of() == pt.TealType.bytes

    cond1, _ = args[0].__teal__(options)
    pred1, pred1End = pt.Seq(args[1]).__teal__(options)
    cond1Branch = pt.TealConditionalBlock([])
    cond2, _ = args[2].__teal__(options)
    pred2, pred2End = pt.Seq(args[3]).__teal__(options)
    cond2Branch = pt.TealConditionalBlock([])
    end = pt.TealSimpleBlock([])

    cond1.setNextBlock(cond1Branch)
    cond1Branch.setTrueBlock(pred1)
    cond1Branch.setFalseBlock(cond2)
    pred1End.setNextBlock(end)

    cond2.setNextBlock(cond2Branch)
    cond2Branch.setTrueBlock(pred2)
    cond2Branch.setFalseBlock(pt.Err().__teal__(options)[0])
    pred2End.setNextBlock(end)

    expected = cond1

    actual, _ = expr.__teal__(options)

    with pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected



import pytest

import pyteal as pt

options = pt.CompileOptions()


def test_while_compiles():

    i = pt.ScratchVar()
    expr = pt.While(pt.Int(2)).Do(pt.Seq([i.store(pt.Int(0))]))
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()
    expr.__teal__(options)


def test_nested_whiles_compile():
    i = pt.ScratchVar()
    expr = pt.While(pt.Int(2)).Do(
        pt.Seq([pt.While(pt.Int(2)).Do(pt.Seq([i.store(pt.Int(0))]))])
    )
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()


def test_continue_break():
    expr = pt.While(pt.Int(0)).Do(pt.Seq([pt.If(pt.Int(1), pt.Break(), pt.Continue())]))
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()
    expr.__teal__(options)


def test_while():
    i = pt.ScratchVar()
    i.store(pt.Int(0))
    items = [i.load() < pt.Int(2), [i.store(i.load() + pt.Int(1))]]
    expr = pt.While(items[0]).Do(pt.Seq(items[1]))
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected, condEnd = items[0].__teal__(options)
    do, doEnd = pt.Seq(items[1]).__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    end = pt.TealSimpleBlock([])

    expectedBranch.setTrueBlock(do)
    expectedBranch.setFalseBlock(end)
    condEnd.setNextBlock(expectedBranch)
    doEnd.setNextBlock(expected)
    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_while_continue():
    i = pt.ScratchVar()
    i.store(pt.Int(0))
    items = [
        i.load() < pt.Int(2),
        i.store(i.load() + pt.Int(1)),
        pt.If(i.load() == pt.Int(1), pt.Continue()),
    ]
    expr = pt.While(items[0]).Do(pt.Seq(items[1], items[2]))
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    options.enterLoop()

    expected, condEnd = items[0].__teal__(options)
    do, doEnd = pt.Seq([items[1], items[2]]).__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    end = pt.TealSimpleBlock([])

    expectedBranch.setTrueBlock(do)
    expectedBranch.setFalseBlock(end)
    condEnd.setNextBlock(expectedBranch)
    doEnd.setNextBlock(expected)

    _, continueBlocks = options.exitLoop()

    for block in continueBlocks:
        block.setNextBlock(do)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_while_break():
    i = pt.ScratchVar()
    i.store(pt.Int(0))
    items = [
        i.load() < pt.Int(2),
        i.store(i.load() + pt.Int(1)),
        pt.If(i.load() == pt.Int(1), pt.Break()),
    ]
    expr = pt.While(items[0]).Do(pt.Seq(items[1], items[2]))
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    options.enterLoop()

    expected, condEnd = items[0].__teal__(options)
    do, doEnd = pt.Seq([items[1], items[2]]).__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    end = pt.TealSimpleBlock([])

    expectedBranch.setTrueBlock(do)
    expectedBranch.setFalseBlock(end)
    condEnd.setNextBlock(expectedBranch)
    doEnd.setNextBlock(expected)

    breakBlocks, _ = options.exitLoop()

    for block in breakBlocks:
        block.setNextBlock(end)

    actual, _ = expr.__teal__(options)

    assert actual == expected


def test_while_invalid():

    with pytest.raises(TypeError):
        expr = pt.While()

    with pytest.raises(pt.TealCompileError):
        expr = pt.While(pt.Int(2))
        expr.type_of()

    with pytest.raises(pt.TealCompileError):
        expr = pt.While(pt.Int(2))
        expr.__teal__(options)

    with pytest.raises(pt.TealCompileError):
        expr = pt.While(pt.Int(2))
        expr.type_of()

    with pytest.raises(pt.TealCompileError):
        expr = pt.While(pt.Int(2))
        expr.__str__()

    with pytest.raises(pt.TealTypeError):
        expr = pt.While(pt.Int(2)).Do(pt.Int(2))

    with pytest.raises(pt.TealTypeError):
        expr = pt.While(pt.Int(2)).Do(pt.Pop(pt.Int(2)), pt.Int(2))

    with pytest.raises(pt.TealCompileError):
        expr = pt.While(pt.Int(0)).Do(pt.Continue()).Do(pt.Continue())
        expr.__str__()


def test_while_multi():
    i = pt.ScratchVar()
    i.store(pt.Int(0))
    items = [i.load() < pt.Int(2), [pt.Pop(pt.Int(1)), i.store(i.load() + pt.Int(1))]]
    expr = pt.While(items[0]).Do(*items[1])
    assert expr.type_of() == pt.TealType.none
    assert not expr.has_return()

    expected, condEnd = items[0].__teal__(options)
    do, doEnd = pt.Seq(items[1]).__teal__(options)
    expectedBranch = pt.TealConditionalBlock([])
    end = pt.TealSimpleBlock([])

    expectedBranch.setTrueBlock(do)
    expectedBranch.setFalseBlock(end)
    condEnd.setNextBlock(expectedBranch)
    doEnd.setNextBlock(expected)
    actual, _ = expr.__teal__(options)

    assert actual == expected



import pyteal as pt


def test_err():
    expr = pt.Err()
    assert expr.type_of() == pt.TealType.none
    assert expr.has_return()
    expected = pt.TealSimpleBlock([pt.TealOp(expr, pt.Op.err)])
    actual, _ = expr.__teal__(pt.CompileOptions())
    assert actual == expected



from typing import List

from pyteal.ir import TealBlock
from pyteal.errors import TealInternalError


def sortBlocks(start: TealBlock, end: TealBlock) -> List[TealBlock]:
    """Topologically sort the graph which starts with the input TealBlock.

    Args:
        start: The starting point of the graph to sort.

    Returns:
        An ordered list of TealBlocks that is sorted such that every block is guaranteed to appear
        in the list before all of its outgoing blocks.
    """
    S = [start]
    order = []
    visited = set()  # I changed visited to a set to be more efficient
    while len(S) != 0:
        n = S.pop()

        if id(n) in visited:
            continue

        S += n.getOutgoing()

        order.append(n)
        visited.add(id(n))

    endIndex = -1
    for i, block in enumerate(order):
        if block is end:
            endIndex = i
            break

    if endIndex == -1:
        raise TealInternalError("End block not present")

    order.pop(endIndex)
    order.append(end)

    return order



from dataclasses import dataclass
from typing import List
import pyteal as pt
from pyteal import abi
import pytest

from pyteal.errors import TealInputError

options = pt.CompileOptions(version=5)


@dataclass
class TransactionTypeTest:
    ts: abi.TransactionTypeSpec
    t: abi.Transaction
    s: str
    txn_type_enum: pt.Expr | None


TransactionValues: List[TransactionTypeTest] = [
    TransactionTypeTest(abi.TransactionTypeSpec(), abi.Transaction(), "txn", None),
    TransactionTypeTest(
        abi.KeyRegisterTransactionTypeSpec(),
        abi.KeyRegisterTransaction(),
        "keyreg",
        pt.TxnType.KeyRegistration,
    ),
    TransactionTypeTest(
        abi.PaymentTransactionTypeSpec(),
        abi.PaymentTransaction(),
        "pay",
        pt.TxnType.Payment,
    ),
    TransactionTypeTest(
        abi.AssetConfigTransactionTypeSpec(),
        abi.AssetConfigTransaction(),
        "acfg",
        pt.TxnType.AssetConfig,
    ),
    TransactionTypeTest(
        abi.AssetFreezeTransactionTypeSpec(),
        abi.AssetFreezeTransaction(),
        "afrz",
        pt.TxnType.AssetFreeze,
    ),
    TransactionTypeTest(
        abi.AssetTransferTransactionTypeSpec(),
        abi.AssetTransferTransaction(),
        "axfer",
        pt.TxnType.AssetTransfer,
    ),
    TransactionTypeTest(
        abi.ApplicationCallTransactionTypeSpec(),
        abi.ApplicationCallTransaction(),
        "appl",
        pt.TxnType.ApplicationCall,
    ),
]


def test_Transaction_str():
    for tv in TransactionValues:
        assert str(tv.ts) == tv.s


def test_TransactionTypeSpec_is_dynamic():
    for tv in TransactionValues:
        assert not (tv.ts).is_dynamic()


def test_TransactionTypeSpec_new_instance():
    for tv in TransactionValues:
        assert isinstance(tv.ts.new_instance(), abi.Transaction)


def test_TransactionTypeSpec_txn_type_enum():
    for tv in TransactionValues:
        if tv.txn_type_enum is None:
            with pytest.raises(
                pt.TealInternalError,
                match=r"abi.TransactionTypeSpec does not represent a specific transaction type$",
            ):
                tv.ts.txn_type_enum()
        else:
            assert tv.ts.txn_type_enum() is tv.txn_type_enum


def test_TransactionTypeSpec_eq():
    for tv in TransactionValues:
        assert tv.ts == tv.ts

        for otherType in (
            abi.ByteTypeSpec(),
            abi.Uint8TypeSpec(),
            abi.AddressTypeSpec(),
        ):
            assert tv.ts != otherType


def test_Transaction_typespec():
    for tv in TransactionValues:
        assert tv.t.type_spec() == tv.ts


def test_Transaction_decode():
    for tv in TransactionValues:
        with pytest.raises(TealInputError):
            tv.t.decode("")


def test_Transaction_encode():
    for tv in TransactionValues:
        with pytest.raises(TealInputError):
            tv.t.encode()


def test_Transaction_get():
    for tv in TransactionValues:
        expr = tv.t.get()
        assert isinstance(expr, pt.TxnObject)


def test_Transaction__set_index():
    for tv in TransactionValues:
        val_to_set = 2
        expr = tv.t._set_index(val_to_set)

        assert expr.type_of() == pt.TealType.none
        assert expr.has_return() is False

        expected = pt.TealSimpleBlock(
            [
                pt.TealOp(expr, pt.Op.int, val_to_set),
                pt.TealOp(None, pt.Op.store, tv.t.stored_value.slot),
            ]
        )
        actual, _ = expr.__teal__(options)
        actual.addIncoming()
        actual = pt.TealBlock.NormalizeBlocks(actual)

        with pt.TealComponent.Context.ignoreExprEquality():
            assert actual == expected



from pyteal import (
    Bytes,
    EcdsaCurve,
    EcdsaDecompress,
    EcdsaRecover,
    EcdsaVerify,
    Int,
    And,
    Subroutine,
    Sha512_256,
    Mode,
    TealType,
)

from tests.blackbox import (
    Blackbox,
    PyTealDryRunExecutor,
)


def test_verify():
    @Blackbox(input_types=[])
    @Subroutine(TealType.uint64)
    def verify():
        return EcdsaVerify(
            EcdsaCurve.Secp256k1,
            Sha512_256(Bytes("testdata")),
            Bytes(
                "base16",
                "33602297203d2753372cea7794ffe1756a278cbc4907b15a0dd132c9fb82555e",
            ),
            Bytes(
                "base16",
                "20f112126cf3e2eac6e8d4f97a403d21bab07b8dbb77154511bb7b07c0173195",
            ),
            (
                Bytes(
                    "base16",
                    "d6143a58c90c06b594e4414cb788659c2805e0056b1dfceea32c03f59efec517",
                ),
                Bytes(
                    "base16",
                    "00bd2400c479efe5ea556f37e1dc11ccb20f1e642dbfe00ca346fffeae508298",
                ),
            ),
        )

    args = []
    app_result = PyTealDryRunExecutor(verify, Mode.Application).dryrun(
        args, compiler_version=5
    )

    assert app_result.stack_top() == 1, app_result.report(
        args, "stack_top() is not equal to 1, indicating ecdsa verification failed."
    )

    @Blackbox(input_types=[])
    @Subroutine(TealType.uint64)
    def verify_fail():
        return EcdsaVerify(
            EcdsaCurve.Secp256k1,
            Sha512_256(Bytes("testdata")),
            Bytes(
                "base16",
                "13602297203d2753372cea7794ffe1756a278cbc4907b15a0dd132c9fb82555e",
            ),
            Bytes(
                "base16",
                "20f112126cf3e2eac6e8d4f97a403d21bab07b8dbb77154511bb7b07c0173195",
            ),
            (
                Bytes(
                    "base16",
                    "d6143a58c90c06b594e4414cb788659c2805e0056b1dfceea32c03f59efec517",
                ),
                Bytes(
                    "base16",
                    "00bd2400c479efe5ea556f37e1dc11ccb20f1e642dbfe00ca346fffeae508298",
                ),
            ),
        )

    args = []
    app_result = PyTealDryRunExecutor(verify_fail, Mode.Application).dryrun(
        args, compiler_version=5
    )

    assert app_result.stack_top() == 0, app_result.report(
        args,
        "stack_top() is not equal to 0, indicating ecdsa verification succeeded when a failure was expected.",
    )


def test_decompress():
    @Blackbox(input_types=[])
    @Subroutine(TealType.uint64)
    def decompress():
        return EcdsaDecompress(
            EcdsaCurve.Secp256k1,
            Bytes(
                "base16",
                "03bd83d54f6a799d05b496653b64bc933e17a898cda4793fe662d50645ecc977d1",
            ),
        ).outputReducer(
            lambda x, y: And(
                x
                == Bytes(
                    "base16",
                    "bd83d54f6a799d05b496653b64bc933e17a898cda4793fe662d50645ecc977d1",
                ),
                y
                == Bytes(
                    "base16",
                    "d4f3063a1ffca4139ea921b5696a6597640289175afece3bc38217a29d6270f9",
                ),
            )
        )

    args = []
    app_result = PyTealDryRunExecutor(decompress, Mode.Application).dryrun(
        args, compiler_version=5
    )

    assert app_result.stack_top() == 1, app_result.report(
        args, "stack_top() is not equal to 1, indicating ecdsa verification failed."
    )


def test_recover():
    @Blackbox(input_types=[])
    @Subroutine(TealType.uint64)
    def recover():
        return EcdsaRecover(
            EcdsaCurve.Secp256k1,
            Sha512_256(Bytes("testdata")),
            Int(1),
            Bytes(
                "base16",
                "cabed943e1403fb93b388174c59a52c759b321855f2d7c4fcc23c99a8a6dce79",
            ),
            Bytes(
                "base16",
                "56192820dde344c32f81450db05e51c6a6f45a2a2db229f657d2c040baf31537",
            ),
        ).outputReducer(
            lambda x, y: And(
                x
                == Bytes(
                    "base16",
                    "71539e0c7a6902a3f5413d6e28a455b2a14316fcf0f6b21193343b3b9d455053",
                ),
                y
                == Bytes(
                    "base16",
                    "fa49ccd95795c7c9a447fdeee83a2193472507a4e41a47e0d50eeeb547b74c51",
                ),
            )
        )

    args = []
    app_result = PyTealDryRunExecutor(recover, Mode.Application).dryrun(
        args, compiler_version=5
    )

    assert app_result.stack_top() == 1, app_result.report(
        args, "stack_top() is not equal to 1, indicating ecdsa verification failed."
    )



# This example is provided for informational purposes only and has not been audited for security.

from pyteal import *


def approval_program():
    on_creation = Seq(
        [
            Assert(Txn.application_args.length() == Int(1)),
            App.globalPut(Bytes("total supply"), Btoi(Txn.application_args[0])),
            App.globalPut(Bytes("reserve"), Btoi(Txn.application_args[0])),
            App.globalPut(Bytes("paused"), Int(0)),
            App.localPut(Int(0), Bytes("contract admin"), Int(1)),
            App.localPut(Int(0), Bytes("transfer admin"), Int(1)),
            App.localPut(Int(0), Bytes("balance"), Int(0)),
            Return(Int(1)),
        ]
    )

    is_contract_admin = App.localGet(Int(0), Bytes("contract admin"))
    is_transfer_admin = App.localGet(Int(0), Bytes("transfer admin"))
    is_any_admin = is_contract_admin.Or(is_transfer_admin)

    can_delete = And(
        is_contract_admin,
        App.globalGet(Bytes("total supply")) == App.globalGet(Bytes("reserve")),
    )

    on_closeout = Seq(
        [
            App.globalPut(
                Bytes("reserve"),
                App.globalGet(Bytes("reserve"))
                + App.localGet(Int(0), Bytes("balance")),
            ),
            Return(Int(1)),
        ]
    )

    register = Seq([App.localPut(Int(0), Bytes("balance"), Int(0)), Return(Int(1))])

    # pause all transfers
    # sender must be any admin
    new_pause_value = Btoi(Txn.application_args[1])
    pause = Seq(
        [
            Assert(Txn.application_args.length() == Int(2)),
            App.globalPut(Bytes("paused"), new_pause_value),
            Return(is_any_admin),
        ]
    )

    # configure the admin status of the account Txn.accounts[1]
    # sender must be contract admin
    new_admin_type = Txn.application_args[1]
    new_admin_status = Btoi(Txn.application_args[2])
    set_admin = Seq(
        [
            Assert(
                And(
                    is_contract_admin,
                    Txn.application_args.length() == Int(3),
                    Or(
                        new_admin_type == Bytes("contract admin"),
                        new_admin_type == Bytes("transfer admin"),
                    ),
                    Txn.accounts.length() == Int(1),
                )
            ),
            App.localPut(Int(1), new_admin_type, new_admin_status),
            Return(Int(1)),
        ]
    )
    # NOTE: The above set_admin code is carefully constructed. If instead we used the following code:
    # Seq([
    #     Assert(And(
    #         Txn.application_args.length() == Int(3),
    #         Or(new_admin_type == Bytes("contract admin"), new_admin_type == Bytes("transfer admin")),
    #         Txn.accounts.length() == Int(1)
    #     )),
    #     App.localPut(Int(1), new_admin_type, new_admin_status),
    #     Return(is_contract_admin)
    # ])
    # It would be vulnerable to the following attack: a sender passes in their own address as
    # Txn.accounts[1], so then the line App.localPut(Int(1), new_admin_type, new_admin_status)
    # changes the sender's admin status, meaning the final Return(is_contract_admin) can return
    # anything the sender wants. This allows anyone to become an admin!

    # freeze Txn.accounts[1]
    # sender must be any admin
    new_freeze_value = Btoi(Txn.application_args[1])
    freeze = Seq(
        [
            Assert(
                And(
                    Txn.application_args.length() == Int(2),
                    Txn.accounts.length() == Int(1),
                )
            ),
            App.localPut(Int(1), Bytes("frozen"), new_freeze_value),
            Return(is_any_admin),
        ]
    )

    # modify the max balance of Txn.accounts[1]
    # if max_balance_value is 0, will delete the existing max balance limitation on the account
    # sender must be transfer admin
    max_balance_value = Btoi(Txn.application_args[1])
    max_balance = Seq(
        [
            Assert(
                And(
                    Txn.application_args.length() == Int(2),
                    Txn.accounts.length() == Int(1),
                )
            ),
            If(
                max_balance_value == Int(0),
                App.localDel(Int(1), Bytes("max balance")),
                App.localPut(Int(1), Bytes("max balance"), max_balance_value),
            ),
            Return(is_transfer_admin),
        ]
    )

    # lock Txn.accounts[1] until a UNIX timestamp
    # sender must be transfer admin
    lock_until_value = Btoi(Txn.application_args[1])
    lock_until = Seq(
        [
            Assert(
                And(
                    Txn.application_args.length() == Int(2),
                    Txn.accounts.length() == Int(1),
                )
            ),
            If(
                lock_until_value == Int(0),
                App.localDel(Int(1), Bytes("lock until")),
                App.localPut(Int(1), Bytes("lock until"), lock_until_value),
            ),
            Return(is_transfer_admin),
        ]
    )

    set_transfer_group = Seq(
        [
            Assert(
                And(
                    Txn.application_args.length() == Int(3),
                    Txn.accounts.length() == Int(1),
                )
            ),
            App.localPut(
                Int(1), Bytes("transfer group"), Btoi(Txn.application_args[2])
            ),
        ]
    )

    def getRuleKey(sendGroup, receiveGroup):
        return Concat(Bytes("rule"), Itob(sendGroup), Itob(receiveGroup))

    lock_transfer_key = getRuleKey(
        Btoi(Txn.application_args[2]), Btoi(Txn.application_args[3])
    )
    lock_transfer_until = Btoi(Txn.application_args[4])
    lock_transfer_group = Seq(
        [
            Assert(Txn.application_args.length() == Int(5)),
            If(
                lock_transfer_until == Int(0),
                App.globalDel(lock_transfer_key),
                App.globalPut(lock_transfer_key, lock_transfer_until),
            ),
        ]
    )

    # sender must be transfer admin
    transfer_group = Seq(
        [
            Assert(Txn.application_args.length() > Int(2)),
            Cond(
                [Txn.application_args[1] == Bytes("set"), set_transfer_group],
                [Txn.application_args[1] == Bytes("lock"), lock_transfer_group],
            ),
            Return(is_transfer_admin),
        ]
    )

    # move assets from the reserve to Txn.accounts[1]
    # sender must be contract admin
    mint_amount = Btoi(Txn.application_args[1])
    mint = Seq(
        [
            Assert(
                And(
                    Txn.application_args.length() == Int(2),
                    Txn.accounts.length() == Int(1),
                    mint_amount <= App.globalGet(Bytes("reserve")),
                )
            ),
            App.globalPut(
                Bytes("reserve"), App.globalGet(Bytes("reserve")) - mint_amount
            ),
            App.localPut(
                Int(1),
                Bytes("balance"),
                App.localGet(Int(1), Bytes("balance")) + mint_amount,
            ),
            Return(is_contract_admin),
        ]
    )

    # move assets from Txn.accounts[1] to the reserve
    # sender must be contract admin
    burn_amount = Btoi(Txn.application_args[1])
    burn = Seq(
        [
            Assert(
                And(
                    Txn.application_args.length() == Int(2),
                    Txn.accounts.length() == Int(1),
                    burn_amount <= App.localGet(Int(1), Bytes("balance")),
                )
            ),
            App.globalPut(
                Bytes("reserve"), App.globalGet(Bytes("reserve")) + burn_amount
            ),
            App.localPut(
                Int(1),
                Bytes("balance"),
                App.localGet(Int(1), Bytes("balance")) - burn_amount,
            ),
            Return(is_contract_admin),
        ]
    )

    # transfer assets from the sender to Txn.accounts[1]
    transfer_amount = Btoi(Txn.application_args[1])
    receiver_max_balance = App.localGetEx(Int(1), App.id(), Bytes("max balance"))
    transfer = Seq(
        [
            Assert(
                And(
                    Txn.application_args.length() == Int(2),
                    Txn.accounts.length() == Int(1),
                    transfer_amount <= App.localGet(Int(0), Bytes("balance")),
                )
            ),
            receiver_max_balance,
            If(
                Or(
                    App.globalGet(Bytes("paused")),
                    App.localGet(Int(0), Bytes("frozen")),
                    App.localGet(Int(1), Bytes("frozen")),
                    App.localGet(Int(0), Bytes("lock until"))
                    >= Global.latest_timestamp(),
                    App.localGet(Int(1), Bytes("lock until"))
                    >= Global.latest_timestamp(),
                    App.globalGet(
                        getRuleKey(
                            App.localGet(Int(0), Bytes("transfer group")),
                            App.localGet(Int(1), Bytes("transfer group")),
                        )
                    )
                    >= Global.latest_timestamp(),
                    And(
                        receiver_max_balance.hasValue(),
                        receiver_max_balance.value()
                        < App.localGet(Int(1), Bytes("balance")) + transfer_amount,
                    ),
                ),
                Return(Int(0)),
            ),
            App.localPut(
                Int(0),
                Bytes("balance"),
                App.localGet(Int(0), Bytes("balance")) - transfer_amount,
            ),
            App.localPut(
                Int(1),
                Bytes("balance"),
                App.localGet(Int(1), Bytes("balance")) + transfer_amount,
            ),
            Return(Int(1)),
        ]
    )

    program = Cond(
        [Txn.application_id() == Int(0), on_creation],
        [Txn.on_completion() == OnComplete.DeleteApplication, Return(can_delete)],
        [
            Txn.on_completion() == OnComplete.UpdateApplication,
            Return(is_contract_admin),
        ],
        [Txn.on_completion() == OnComplete.CloseOut, on_closeout],
        [Txn.on_completion() == OnComplete.OptIn, register],
        [Txn.application_args[0] == Bytes("pause"), pause],
        [Txn.application_args[0] == Bytes("set admin"), set_admin],
        [Txn.application_args[0] == Bytes("freeze"), freeze],
        [Txn.application_args[0] == Bytes("max balance"), max_balance],
        [Txn.application_args[0] == Bytes("lock until"), lock_until],
        [Txn.application_args[0] == Bytes("transfer group"), transfer_group],
        [Txn.application_args[0] == Bytes("mint"), mint],
        [Txn.application_args[0] == Bytes("burn"), burn],
        [Txn.application_args[0] == Bytes("transfer"), transfer],
    )

    return program


def clear_state_program():
    program = Seq(
        [
            App.globalPut(
                Bytes("reserve"),
                App.globalGet(Bytes("reserve"))
                + App.localGet(Int(0), Bytes("balance")),
            ),
            Return(Int(1)),
        ]
    )

    return program


if __name__ == "__main__":
    with open("security_token_approval.teal", "w") as f:
        compiled = compileTeal(approval_program(), mode=Mode.Application, version=2)
        f.write(compiled)

    with open("security_token_clear_state.teal", "w") as f:
        compiled = compileTeal(clear_state_program(), mode=Mode.Application, version=2)
        f.write(compiled)



#!/usr/bin/env python3

import base64
from nacl import encoding, hash
import params
import re
import time
import uuid

from algosdk import algod
from algosdk.future import transaction

from pyteal import *

from recurring_swap import recurring_swap


# ------- generate provider's account -----------------------------------------------
key_fn = str(uuid.uuid4()) + ".key"
execute(["algokey", "generate", "-f", key_fn])
stdout, stderr = execute(["algokey", "export", "-f", key_fn])
print("generated key file {}".format(key_fn))

if stderr != "":
    print(stderr)
    raise

result = re.search(r"key: \w+", stdout)
provider_addr = result.group(0)[5:]
print("provider addr: {}".format(provider_addr))

# ------- instantiate template, compile teal source, get escrow address -------------

program = recurring_swap(tmpl_provider=Addr(provider_addr))
teal_source = compileTeal(program, mode=Mode.Signature, version=2)
# print(teal_source)

# compile teal
teal_base = str(uuid.uuid4())
teal_file = teal_base + ".teal"
with open(teal_file, "w+") as f:
    f.write(teal_source)
lsig_fname = teal_base + ".tealc"

stdout, stderr = execute(["goal", "clerk", "compile", "-o", lsig_fname, teal_file])

if stderr != "":
    print(stderr)
    raise
elif len(stdout) < 59:
    print("error in compile teal")
    raise

result = re.search(r": \w+", stdout)
escrow_addr = result.group(0)[2:]
print("Dispense at least 202000 microAlgo to {}".format(escrow_addr))
input("Make sure you did that. Press Enter to continue...")

# now, as a provider, you can withdraw Algo from the escrow if you sign the first valid
acl = algod.AlgodClient(params.algod_token, params.algod_address)

sp = acl.suggested_params_as_object()
first_valid = sp.first
data = first_valid.to_bytes(8, byteorder="big")
lease = hash.sha256(data)
lease_bytes = encoding.HexEncoder.decode(lease)
print("first valid: {}".format(first_valid))

txn = transaction.PaymentTxn(escrow_addr, sp, provider_addr, 100000, lease=lease_bytes)

with open(lsig_fname, "rb") as f:
    teal_bytes = f.read()
lsig = transaction.LogicSig(teal_bytes)
lstx = transaction.LogicSigTransaction(txn, lsig)

assert lstx.verify()

# send LogicSigTransaction to network
transaction.write_to_file([lstx], "r_s_1.txn")

stdout, stderr = execute(
    [
        "goal",
        "clerk",
        "tealsign",
        "--data-b64",
        base64.b64encode(data),
        "--lsig-txn",
        "r_s_1.txn",
        "--keyfile",
        key_fn,
        "--set-lsig-arg-idx",
        "0",
    ]
)
if stderr != "":
    print(stderr)
    raise

print(stdout)

lstx = transaction.retrieve_from_file("r_s_1.txn")
txid = acl.send_transactions(lstx)

print("1st withraw Succesfull! txid:{}".format(txid))

# at least sleep to the next round
time.sleep(6)

sp = acl.suggested_params_as_object()
first_valid = sp.first
data = first_valid.to_bytes(8, byteorder="big")
lease = hash.sha256(data)
lease_bytes = encoding.HexEncoder.decode(lease)

print("first valid: {}".format(first_valid))

txn = transaction.PaymentTxn(escrow_addr, sp, provider_addr, 100000, lease=lease_bytes)

with open(lsig_fname, "rb") as f:
    teal_bytes = f.read()
lsig = transaction.LogicSig(teal_bytes)
lstx = transaction.LogicSigTransaction(txn, lsig)

assert lstx.verify()

# send LogicSigTransaction to network
transaction.write_to_file([lstx], "r_s_2.txn")

stdout, stderr = execute(
    [
        "goal",
        "clerk",
        "tealsign",
        "--data-b64",
        base64.b64encode(data),
        "--lsig-txn",
        "r_s_2.txn",
        "--keyfile",
        key_fn,
        "--set-lsig-arg-idx",
        "0",
    ]
)
if stderr != "":
    print(stderr)
    raise

print(stdout)

lstx = transaction.retrieve_from_file("r_s_2.txn")
txid = acl.send_transactions(lstx)
print("2nd withraw Succesfull! txid:{}".format(txid))



import pytest
from typing import NamedTuple

import pyteal as pt

avm6Options = pt.CompileOptions(version=6)


def test_OpUp_init():
    app_id = pt.Int(1)
    opup_explicit = pt.OpUp(pt.OpUpMode.Explicit, target_app_id=app_id)
    assert opup_explicit.mode is pt.OpUpMode.Explicit
    assert opup_explicit.target_app_id is app_id

    with pytest.raises(
        pt.TealInputError, match="target_app_id must be specified in Explicit OpUp mode"
    ):
        pt.OpUp(pt.OpUpMode.Explicit)

    with pytest.raises(pt.TealTypeError):
        pt.OpUp(pt.OpUpMode.Explicit, target_app_id=pt.Bytes("appid"))

    opup_oncall = pt.OpUp(pt.OpUpMode.OnCall)
    assert opup_oncall.mode is pt.OpUpMode.OnCall
    assert opup_oncall.target_app_id is None

    with pytest.raises(
        pt.TealInputError, match="target_app_id is not used in OnCall OpUp mode"
    ):
        pt.OpUp(pt.OpUpMode.OnCall, target_app_id=app_id)

    with pytest.raises(pt.TealInputError, match="Invalid OpUp mode provided"):
        pt.OpUp(None)


class OpUpTest(NamedTuple):
    opup: pt.OpUp
    fee_source: pt.OpUpFeeSource | None
    expected_inner_fields: dict[pt.TxnField, pt.Expr | list[pt.Expr]]


OP_UP_TEST_CASES: list[OpUpTest] = [
    OpUpTest(
        opup=pt.OpUp(pt.OpUpMode.Explicit, target_app_id=pt.Int(9)),
        fee_source=None,
        expected_inner_fields={
            pt.TxnField.type_enum: pt.TxnType.ApplicationCall,
            pt.TxnField.application_id: pt.Int(9),
        },
    ),
    OpUpTest(
        opup=pt.OpUp(pt.OpUpMode.Explicit, target_app_id=pt.Int(876)),
        fee_source=pt.OpUpFeeSource.Any,
        expected_inner_fields={
            pt.TxnField.type_enum: pt.TxnType.ApplicationCall,
            pt.TxnField.application_id: pt.Int(876),
        },
    ),
    OpUpTest(
        opup=pt.OpUp(pt.OpUpMode.Explicit, target_app_id=pt.Int(1000001)),
        fee_source=pt.OpUpFeeSource.GroupCredit,
        expected_inner_fields={
            pt.TxnField.type_enum: pt.TxnType.ApplicationCall,
            pt.TxnField.fee: pt.Int(0),
            pt.TxnField.application_id: pt.Int(1000001),
        },
    ),
    OpUpTest(
        opup=pt.OpUp(pt.OpUpMode.Explicit, target_app_id=pt.Int(3)),
        fee_source=pt.OpUpFeeSource.AppAccount,
        expected_inner_fields={
            pt.TxnField.type_enum: pt.TxnType.ApplicationCall,
            pt.TxnField.fee: pt.Global.min_txn_fee(),
            pt.TxnField.application_id: pt.Int(3),
        },
    ),
    OpUpTest(
        opup=pt.OpUp(pt.OpUpMode.OnCall),
        fee_source=None,
        expected_inner_fields={
            pt.TxnField.type_enum: pt.TxnType.ApplicationCall,
            pt.TxnField.on_completion: pt.OnComplete.DeleteApplication,
            pt.TxnField.approval_program: pt.Bytes("base16", "068101"),
            pt.TxnField.clear_state_program: pt.Bytes("base16", "068101"),
        },
    ),
    OpUpTest(
        opup=pt.OpUp(pt.OpUpMode.OnCall),
        fee_source=pt.OpUpFeeSource.Any,
        expected_inner_fields={
            pt.TxnField.type_enum: pt.TxnType.ApplicationCall,
            pt.TxnField.on_completion: pt.OnComplete.DeleteApplication,
            pt.TxnField.approval_program: pt.Bytes("base16", "068101"),
            pt.TxnField.clear_state_program: pt.Bytes("base16", "068101"),
        },
    ),
    OpUpTest(
        opup=pt.OpUp(pt.OpUpMode.OnCall),
        fee_source=pt.OpUpFeeSource.GroupCredit,
        expected_inner_fields={
            pt.TxnField.type_enum: pt.TxnType.ApplicationCall,
            pt.TxnField.fee: pt.Int(0),
            pt.TxnField.on_completion: pt.OnComplete.DeleteApplication,
            pt.TxnField.approval_program: pt.Bytes("base16", "068101"),
            pt.TxnField.clear_state_program: pt.Bytes("base16", "068101"),
        },
    ),
    OpUpTest(
        opup=pt.OpUp(pt.OpUpMode.OnCall),
        fee_source=pt.OpUpFeeSource.AppAccount,
        expected_inner_fields={
            pt.TxnField.type_enum: pt.TxnType.ApplicationCall,
            pt.TxnField.fee: pt.Global.min_txn_fee(),
            pt.TxnField.on_completion: pt.OnComplete.DeleteApplication,
            pt.TxnField.approval_program: pt.Bytes("base16", "068101"),
            pt.TxnField.clear_state_program: pt.Bytes("base16", "068101"),
        },
    ),
]


def test_OP_UP_TEST_CASES_is_exhaustive():
    fee_sources: list[pt.OpUpFeeSource | None] = [fs for fs in pt.OpUpFeeSource]
    fee_sources.append(None)

    for mode, fee_source in zip(pt.OpUpMode, fee_sources):
        found_combination = False

        for test_case in OP_UP_TEST_CASES:
            if test_case.opup.mode is mode and test_case.fee_source is fee_source:
                found_combination = True
                break

        assert (
            found_combination
        ), f"Combination not found in test cases: mode={mode}, fee_source={fee_source}"


@pytest.mark.parametrize("opup, fee_source, expected_inner_fields", OP_UP_TEST_CASES)
def test_OpUp_ensure_budget(
    opup: pt.OpUp,
    fee_source: pt.OpUpFeeSource | None,
    expected_inner_fields: dict[pt.TxnField, pt.Expr | list[pt.Expr]],
):
    required_budget = pt.Int(777)
    expr = (
        opup.ensure_budget(required_budget)
        if fee_source is None
        else opup.ensure_budget(required_budget, fee_source)
    )
    assert expr.type_of() == pt.TealType.none
    assert expr.has_return() is False

    intermediate_value = pt.ScratchVar()
    expected_expr = pt.Seq(
        intermediate_value.store(required_budget + pt.Int(10)),
        pt.While(intermediate_value.load() > pt.Global.opcode_budget()).Do(
            pt.InnerTxnBuilder.Execute(expected_inner_fields)
        ),
    )
    expected, _ = expected_expr.__teal__(avm6Options)
    expected.addIncoming()
    expected = pt.TealBlock.NormalizeBlocks(expected)

    actual, _ = expr.__teal__(avm6Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreScratchSlotEquality(), pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    assert pt.TealBlock.MatchScratchSlotReferences(
        pt.TealBlock.GetReferencedScratchSlots(actual),
        pt.TealBlock.GetReferencedScratchSlots(expected),
    )

    with pytest.raises(pt.TealTypeError):
        opup.ensure_budget(required_budget=pt.Bytes("budget"))

    with pytest.raises(pt.TealTypeError):
        opup.ensure_budget(
            required_budget=required_budget, fee_source=pt.Bytes("fee_src")  # type: ignore[arg-type]
        )


@pytest.mark.parametrize("opup, fee_source, expected_inner_fields", OP_UP_TEST_CASES)
def test_OpUp_maximize_budget(
    opup: pt.OpUp,
    fee_source: pt.OpUpFeeSource | None,
    expected_inner_fields: dict[pt.TxnField, pt.Expr | list[pt.Expr]],
):
    fee = pt.Int(12345)
    expr = (
        opup.maximize_budget(fee)
        if fee_source is None
        else opup.maximize_budget(fee, fee_source)
    )
    assert expr.type_of() == pt.TealType.none
    assert expr.has_return() is False

    intermediate_value = pt.ScratchVar()
    expected_expr = pt.For(
        intermediate_value.store(pt.Int(0)),
        intermediate_value.load() < fee / pt.Global.min_txn_fee(),
        intermediate_value.store(intermediate_value.load() + pt.Int(1)),
    ).Do(pt.InnerTxnBuilder.Execute(expected_inner_fields))
    expected, _ = expected_expr.__teal__(avm6Options)
    expected.addIncoming()
    expected = pt.TealBlock.NormalizeBlocks(expected)

    actual, _ = expr.__teal__(avm6Options)
    actual.addIncoming()
    actual = pt.TealBlock.NormalizeBlocks(actual)

    with pt.TealComponent.Context.ignoreScratchSlotEquality(), pt.TealComponent.Context.ignoreExprEquality():
        assert actual == expected

    assert pt.TealBlock.MatchScratchSlotReferences(
        pt.TealBlock.GetReferencedScratchSlots(actual),
        pt.TealBlock.GetReferencedScratchSlots(expected),
    )

    with pytest.raises(pt.TealTypeError):
        opup.maximize_budget(fee=pt.Bytes("fee"))

    with pytest.raises(pt.TealTypeError):
        opup.maximize_budget(fee=fee, fee_source=pt.Bytes("fee_src"))  # type: ignore[arg-type]



from typing import TYPE_CHECKING

from pyteal.types import TealType, require_type, types_match
from pyteal.errors import verifyProgramVersion, TealCompileError
from pyteal.ir import TealOp, Op, TealBlock
from pyteal.ast.expr import Expr
from pyteal.ast.int import Int

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class Return(Expr):
    """Return a value from the current execution context."""

    def __init__(self, value: Expr = None) -> None:
        """Create a new Return expression.

        If called from the main program, this will immediately exit the program
        and the value returned will be the program's success value (must be a
        uint64, 0 indicates failure, 1 or greater indicates success).

        If called from within a subroutine, this will return from the current
        subroutine with either no value if the subroutine does not produce a
        return value, or the given return value if it does produce a return value.
        """
        super().__init__()
        if value is not None:
            require_type(value, TealType.anytype)
        self.value = value

    def __teal__(self, options: "CompileOptions"):
        if options.currentSubroutine is not None:
            verifyProgramVersion(
                Op.retsub.min_version,
                options.version,
                "Program version too low to use subroutines",
            )
            returnType = options.currentSubroutine.return_type
            if returnType == TealType.none:
                if self.value is not None:
                    raise TealCompileError(
                        "Cannot return a value from a subroutine with return type TealType.none",
                        self,
                    )
            else:
                if self.value is None:
                    raise TealCompileError(
                        "A subroutine declares it returns a value, but no value is being returned",
                        self,
                    )
                actualType = self.value.type_of()
                if not types_match(actualType, returnType):
                    raise TealCompileError(
                        "Incompatible return type from subroutine, expected {} but got {}".format(
                            returnType, actualType
                        ),
                        self,
                    )
            op = Op.retsub
        else:
            if self.value is None:
                raise TealCompileError(
                    "Return from main program must have an argument", self
                )
            actualType = self.value.type_of()
            if not types_match(actualType, TealType.uint64):
                raise TealCompileError(
                    "Incompatible return type from main program, expected {} but got {}".format(
                        TealType.uint64, actualType
                    ),
                    self,
                )
            op = Op.return_

        args = [] if self.value is None else [self.value]
        return TealBlock.FromOp(options, TealOp(self, op), *args)

    def __str__(self):
        return "(Return {})".format(self.value)

    def type_of(self):
        return TealType.none

    def has_return(self):
        return True


Return.__module__ = "pyteal"


class ExitProgram(Expr):
    """Immediately exit the program with the indicated success value."""

    def __init__(self, success: Expr) -> None:
        super().__init__()
        require_type(success, TealType.uint64)
        self.success = success

    def __teal__(self, options: "CompileOptions"):
        return TealBlock.FromOp(options, TealOp(self, Op.return_), self.success)

    def __str__(self):
        return "(ExitProgram {})".format(self.success)

    def type_of(self):
        return TealType.none

    def has_return(self):
        return True


ExitProgram.__module__ = "pyteal"


def Approve() -> Expr:
    """Immediately exit the program and mark the execution as successful."""
    return ExitProgram(Int(1))


def Reject() -> Expr:
    """Immediately exit the program and mark the execution as unsuccessful."""
    return ExitProgram(Int(0))



from collections import OrderedDict

import pytest

import pyteal as pt

from pyteal.compiler.subroutines import (
    findRecursionPoints,
    spillLocalSlotsDuringRecursion,
    resolveSubroutines,
)


def test_findRecursionPoints_empty():
    subroutines = dict()

    expected = dict()
    actual = findRecursionPoints(subroutines)
    assert actual == expected


def test_findRecursionPoints_none():
    def sub1Impl():
        return None

    def sub2Impl(a1):
        return None

    def sub3Impl(a1, a2, a3):
        return None

    subroutine1 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
    subroutine2 = pt.SubroutineDefinition(sub2Impl, pt.TealType.bytes)
    subroutine3 = pt.SubroutineDefinition(sub3Impl, pt.TealType.none)

    subroutines = {
        subroutine1: {subroutine2, subroutine3},
        subroutine2: {subroutine3},
        subroutine3: set(),
    }

    expected = {
        subroutine1: set(),
        subroutine2: set(),
        subroutine3: set(),
    }

    actual = findRecursionPoints(subroutines)
    assert actual == expected


def test_findRecursionPoints_direct_recursion():
    def sub1Impl():
        return None

    def sub2Impl(a1):
        return None

    def sub3Impl(a1, a2, a3):
        return None

    subroutine1 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
    subroutine2 = pt.SubroutineDefinition(sub2Impl, pt.TealType.bytes)
    subroutine3 = pt.SubroutineDefinition(sub3Impl, pt.TealType.none)

    subroutines = {
        subroutine1: {subroutine2, subroutine3},
        subroutine2: {subroutine2, subroutine3},
        subroutine3: set(),
    }

    expected = {
        subroutine1: set(),
        subroutine2: {subroutine2},
        subroutine3: set(),
    }

    actual = findRecursionPoints(subroutines)
    assert actual == expected


def test_findRecursionPoints_mutual_recursion():
    def sub1Impl():
        return None

    def sub2Impl(a1):
        return None

    def sub3Impl(a1, a2, a3):
        return None

    subroutine1 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
    subroutine2 = pt.SubroutineDefinition(sub2Impl, pt.TealType.bytes)
    subroutine3 = pt.SubroutineDefinition(sub3Impl, pt.TealType.none)

    subroutines = {
        subroutine1: {subroutine2, subroutine3},
        subroutine2: {subroutine1, subroutine3},
        subroutine3: set(),
    }

    expected = {
        subroutine1: {subroutine2},
        subroutine2: {subroutine1},
        subroutine3: set(),
    }

    actual = findRecursionPoints(subroutines)
    assert actual == expected


def test_findRecursionPoints_direct_and_mutual_recursion():
    def sub1Impl():
        return None

    def sub2Impl(a1):
        return None

    def sub3Impl(a1, a2, a3):
        return None

    subroutine1 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
    subroutine2 = pt.SubroutineDefinition(sub2Impl, pt.TealType.bytes)
    subroutine3 = pt.SubroutineDefinition(sub3Impl, pt.TealType.none)

    subroutines = {
        subroutine1: {subroutine2, subroutine3},
        subroutine2: {subroutine1, subroutine2, subroutine3},
        subroutine3: set(),
    }

    expected = {
        subroutine1: {subroutine2},
        subroutine2: {subroutine1, subroutine2},
        subroutine3: set(),
    }

    actual = findRecursionPoints(subroutines)
    assert actual == expected


def test_spillLocalSlotsDuringRecursion_no_subroutines():
    for version in (4, 5):
        l1Label = pt.LabelReference("l1")
        mainOps = [
            pt.TealOp(None, pt.Op.txn, "Fee"),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bz, l1Label),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.return_),
            pt.TealLabel(None, l1Label),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.return_),
        ]

        subroutineMapping = {None: mainOps}

        subroutineGraph = dict()

        localSlots = {None: set()}

        spillLocalSlotsDuringRecursion(
            version, subroutineMapping, subroutineGraph, localSlots
        )

        assert mainOps == [
            pt.TealOp(None, pt.Op.txn, "Fee"),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bz, l1Label),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.return_),
            pt.TealLabel(None, l1Label),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.return_),
        ]


def test_spillLocalSlotsDuringRecursion_1_subroutine_no_recursion():
    for version in (4, 5):
        subroutine = pt.SubroutineDefinition(lambda: None, pt.TealType.uint64)

        subroutineL1Label = pt.LabelReference("l1")
        subroutineOps = [
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bnz, subroutineL1Label),
            pt.TealOp(None, pt.Op.err),
            pt.TealLabel(None, subroutineL1Label),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.retsub),
        ]

        l1Label = pt.LabelReference("l1")
        mainOps = [
            pt.TealOp(None, pt.Op.txn, "Fee"),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bz, l1Label),
            pt.TealOp(None, pt.Op.int, 100),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.return_),
            pt.TealLabel(None, l1Label),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.return_),
        ]

        subroutineMapping = {None: mainOps, subroutine: subroutineOps}

        subroutineGraph = {subroutine: set()}

        localSlots = {None: set(), subroutine: {0}}

        spillLocalSlotsDuringRecursion(
            version, subroutineMapping, subroutineGraph, localSlots
        )

        assert subroutineMapping == {
            None: [
                pt.TealOp(None, pt.Op.txn, "Fee"),
                pt.TealOp(None, pt.Op.int, 0),
                pt.TealOp(None, pt.Op.eq),
                pt.TealOp(None, pt.Op.bz, l1Label),
                pt.TealOp(None, pt.Op.int, 100),
                pt.TealOp(None, pt.Op.callsub, subroutine),
                pt.TealOp(None, pt.Op.return_),
                pt.TealLabel(None, l1Label),
                pt.TealOp(None, pt.Op.int, 0),
                pt.TealOp(None, pt.Op.return_),
            ],
            subroutine: [
                pt.TealOp(None, pt.Op.store, 0),
                pt.TealOp(None, pt.Op.load, 0),
                pt.TealOp(None, pt.Op.int, 0),
                pt.TealOp(None, pt.Op.eq),
                pt.TealOp(None, pt.Op.bnz, subroutineL1Label),
                pt.TealOp(None, pt.Op.err),
                pt.TealLabel(None, subroutineL1Label),
                pt.TealOp(None, pt.Op.int, 1),
                pt.TealOp(None, pt.Op.retsub),
            ],
        }


def test_spillLocalSlotsDuringRecursion_1_subroutine_recursion_v4():
    def sub1Impl(a1):
        return None

    subroutine = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)

    subroutineL1Label = pt.LabelReference("l1")
    subroutineOps = [
        pt.TealOp(None, pt.Op.store, 0),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bnz, subroutineL1Label),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.minus),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.retsub),
        pt.TealLabel(None, subroutineL1Label),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    l1Label = pt.LabelReference("l1")
    mainOps = [
        pt.TealOp(None, pt.Op.txn, "Fee"),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bz, l1Label),
        pt.TealOp(None, pt.Op.int, 100),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.return_),
        pt.TealLabel(None, l1Label),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.return_),
    ]

    subroutineMapping = {None: mainOps, subroutine: subroutineOps}

    subroutineGraph = {subroutine: {subroutine}}

    localSlots = {None: set(), subroutine: {0}}

    spillLocalSlotsDuringRecursion(4, subroutineMapping, subroutineGraph, localSlots)

    assert subroutineMapping == {
        None: [
            pt.TealOp(None, pt.Op.txn, "Fee"),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bz, l1Label),
            pt.TealOp(None, pt.Op.int, 100),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.return_),
            pt.TealLabel(None, l1Label),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.return_),
        ],
        subroutine: [
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bnz, subroutineL1Label),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.minus),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.dig, 1),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.swap),
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.swap),
            pt.TealOp(None, pt.Op.pop),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.add),
            pt.TealOp(None, pt.Op.retsub),
            pt.TealLabel(None, subroutineL1Label),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.retsub),
        ],
    }


def test_spillLocalSlotsDuringRecursion_1_subroutine_recursion_v5():
    def sub1Impl(a1):
        return None

    subroutine = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)

    subroutineL1Label = pt.LabelReference("l1")
    subroutineOps = [
        pt.TealOp(None, pt.Op.store, 0),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bnz, subroutineL1Label),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.minus),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.retsub),
        pt.TealLabel(None, subroutineL1Label),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    l1Label = pt.LabelReference("l1")
    mainOps = [
        pt.TealOp(None, pt.Op.txn, "Fee"),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bz, l1Label),
        pt.TealOp(None, pt.Op.int, 100),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.return_),
        pt.TealLabel(None, l1Label),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.return_),
    ]

    subroutineMapping = {None: mainOps, subroutine: subroutineOps}

    subroutineGraph = {subroutine: {subroutine}}

    localSlots = {None: set(), subroutine: {0}}

    spillLocalSlotsDuringRecursion(5, subroutineMapping, subroutineGraph, localSlots)

    assert subroutineMapping == {
        None: [
            pt.TealOp(None, pt.Op.txn, "Fee"),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bz, l1Label),
            pt.TealOp(None, pt.Op.int, 100),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.return_),
            pt.TealLabel(None, l1Label),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.return_),
        ],
        subroutine: [
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bnz, subroutineL1Label),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.minus),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.swap),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.swap),
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.add),
            pt.TealOp(None, pt.Op.retsub),
            pt.TealLabel(None, subroutineL1Label),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.retsub),
        ],
    }


def test_spillLocalSlotsDuringRecursion_multiple_subroutines_no_recursion():
    for version in (4, 5):

        def sub1Impl(a1):
            return None

        def sub2Impl(a1, a2):
            return None

        def sub3Impl(a1, a2, a3):
            return None

        subroutine1 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
        subroutine2 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
        subroutine3 = pt.SubroutineDefinition(sub1Impl, pt.TealType.none)

        subroutine1L1Label = pt.LabelReference("l1")
        subroutine1Ops = [
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bnz, subroutine1L1Label),
            pt.TealOp(None, pt.Op.err),
            pt.TealLabel(None, subroutine1L1Label),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.callsub, subroutine3),
            pt.TealOp(None, pt.Op.retsub),
        ]

        subroutine2L1Label = pt.LabelReference("l1")
        subroutine2Ops = [
            pt.TealOp(None, pt.Op.store, 1),
            pt.TealOp(None, pt.Op.load, 1),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bnz, subroutine2L1Label),
            pt.TealOp(None, pt.Op.err),
            pt.TealLabel(None, subroutine2L1Label),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.retsub),
        ]

        subroutine3Ops = [
            pt.TealOp(None, pt.Op.retsub),
        ]

        l1Label = pt.LabelReference("l1")
        mainOps = [
            pt.TealOp(None, pt.Op.txn, "Fee"),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bz, l1Label),
            pt.TealOp(None, pt.Op.int, 100),
            pt.TealOp(None, pt.Op.callsub, subroutine1),
            pt.TealOp(None, pt.Op.return_),
            pt.TealLabel(None, l1Label),
            pt.TealOp(None, pt.Op.int, 101),
            pt.TealOp(None, pt.Op.callsub, subroutine2),
            pt.TealOp(None, pt.Op.return_),
        ]

        subroutineMapping = {
            None: mainOps,
            subroutine1: subroutine1Ops,
            subroutine2: subroutine2Ops,
            subroutine3: subroutine3Ops,
        }

        subroutineGraph = {
            subroutine1: {subroutine2},
            subroutine2: set(),
            subroutine3: set(),
        }

        localSlots = {None: set(), subroutine1: {0}, subroutine2: {1}, subroutine3: {}}

        spillLocalSlotsDuringRecursion(
            version, subroutineMapping, subroutineGraph, localSlots
        )

        assert subroutineMapping == {
            None: [
                pt.TealOp(None, pt.Op.txn, "Fee"),
                pt.TealOp(None, pt.Op.int, 0),
                pt.TealOp(None, pt.Op.eq),
                pt.TealOp(None, pt.Op.bz, l1Label),
                pt.TealOp(None, pt.Op.int, 100),
                pt.TealOp(None, pt.Op.callsub, subroutine1),
                pt.TealOp(None, pt.Op.return_),
                pt.TealLabel(None, l1Label),
                pt.TealOp(None, pt.Op.int, 101),
                pt.TealOp(None, pt.Op.callsub, subroutine2),
                pt.TealOp(None, pt.Op.return_),
            ],
            subroutine1: [
                pt.TealOp(None, pt.Op.store, 0),
                pt.TealOp(None, pt.Op.load, 0),
                pt.TealOp(None, pt.Op.int, 0),
                pt.TealOp(None, pt.Op.eq),
                pt.TealOp(None, pt.Op.bnz, subroutine1L1Label),
                pt.TealOp(None, pt.Op.err),
                pt.TealLabel(None, subroutine1L1Label),
                pt.TealOp(None, pt.Op.int, 1),
                pt.TealOp(None, pt.Op.callsub, subroutine3),
                pt.TealOp(None, pt.Op.retsub),
            ],
            subroutine2: [
                pt.TealOp(None, pt.Op.store, 1),
                pt.TealOp(None, pt.Op.load, 1),
                pt.TealOp(None, pt.Op.int, 0),
                pt.TealOp(None, pt.Op.eq),
                pt.TealOp(None, pt.Op.bnz, subroutine2L1Label),
                pt.TealOp(None, pt.Op.err),
                pt.TealLabel(None, subroutine2L1Label),
                pt.TealOp(None, pt.Op.int, 1),
                pt.TealOp(None, pt.Op.retsub),
            ],
            subroutine3: [
                pt.TealOp(None, pt.Op.retsub),
            ],
        }


def test_spillLocalSlotsDuringRecursion_multiple_subroutines_recursion_v4():
    def sub1Impl(a1):
        return None

    def sub2Impl(a1, a2):
        return None

    def sub3Impl(a1, a2, a3):
        return None

    subroutine1 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
    subroutine2 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
    subroutine3 = pt.SubroutineDefinition(sub1Impl, pt.TealType.none)

    subroutine1L1Label = pt.LabelReference("l1")
    subroutine1Ops = [
        pt.TealOp(None, pt.Op.store, 0),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bnz, subroutine1L1Label),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.minus),
        pt.TealOp(None, pt.Op.callsub, subroutine1),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.retsub),
        pt.TealLabel(None, subroutine1L1Label),
        pt.TealOp(None, pt.Op.load, 255),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine2L1Label = pt.LabelReference("l1")
    subroutine2Ops = [
        pt.TealOp(None, pt.Op.store, 1),
        pt.TealOp(None, pt.Op.load, 1),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bnz, subroutine2L1Label),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.minus),
        pt.TealOp(None, pt.Op.callsub, subroutine1),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.retsub),
        pt.TealLabel(None, subroutine2L1Label),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine3Ops = [
        pt.TealOp(None, pt.Op.callsub, subroutine3),
        pt.TealOp(None, pt.Op.retsub),
    ]

    l1Label = pt.LabelReference("l1")
    mainOps = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, 255),
        pt.TealOp(None, pt.Op.txn, "Fee"),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bz, l1Label),
        pt.TealOp(None, pt.Op.int, 100),
        pt.TealOp(None, pt.Op.callsub, subroutine1),
        pt.TealOp(None, pt.Op.return_),
        pt.TealLabel(None, l1Label),
        pt.TealOp(None, pt.Op.int, 101),
        pt.TealOp(None, pt.Op.callsub, subroutine2),
        pt.TealOp(None, pt.Op.return_),
        pt.TealOp(None, pt.Op.callsub, subroutine3),
    ]

    subroutineMapping = {
        None: mainOps,
        subroutine1: subroutine1Ops,
        subroutine2: subroutine2Ops,
        subroutine3: subroutine3Ops,
    }

    subroutineGraph = {
        subroutine1: {subroutine1},
        subroutine2: {subroutine1},
        subroutine3: {subroutine3},
    }

    localSlots = {None: set(), subroutine1: {0}, subroutine2: {1}, subroutine3: {}}

    spillLocalSlotsDuringRecursion(4, subroutineMapping, subroutineGraph, localSlots)

    assert subroutineMapping == {
        None: [
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.store, 255),
            pt.TealOp(None, pt.Op.txn, "Fee"),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bz, l1Label),
            pt.TealOp(None, pt.Op.int, 100),
            pt.TealOp(None, pt.Op.callsub, subroutine1),
            pt.TealOp(None, pt.Op.return_),
            pt.TealLabel(None, l1Label),
            pt.TealOp(None, pt.Op.int, 101),
            pt.TealOp(None, pt.Op.callsub, subroutine2),
            pt.TealOp(None, pt.Op.return_),
            pt.TealOp(None, pt.Op.callsub, subroutine3),
        ],
        subroutine1: [
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bnz, subroutine1L1Label),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.minus),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.dig, 1),
            pt.TealOp(None, pt.Op.callsub, subroutine1),
            pt.TealOp(None, pt.Op.swap),
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.swap),
            pt.TealOp(None, pt.Op.pop),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.add),
            pt.TealOp(None, pt.Op.retsub),
            pt.TealLabel(None, subroutine1L1Label),
            pt.TealOp(None, pt.Op.load, 255),
            pt.TealOp(None, pt.Op.retsub),
        ],
        subroutine2: [
            pt.TealOp(None, pt.Op.store, 1),
            pt.TealOp(None, pt.Op.load, 1),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bnz, subroutine2L1Label),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.minus),
            pt.TealOp(None, pt.Op.callsub, subroutine1),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.add),
            pt.TealOp(None, pt.Op.retsub),
            pt.TealLabel(None, subroutine2L1Label),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.retsub),
        ],
        subroutine3: [
            pt.TealOp(None, pt.Op.callsub, subroutine3),
            pt.TealOp(None, pt.Op.retsub),
        ],
    }


def test_spillLocalSlotsDuringRecursion_multiple_subroutines_recursion_v5():
    def sub1Impl(a1):
        return None

    def sub2Impl(a1, a2):
        return None

    def sub3Impl(a1, a2, a3):
        return None

    subroutine1 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
    subroutine2 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
    subroutine3 = pt.SubroutineDefinition(sub1Impl, pt.TealType.none)

    subroutine1L1Label = pt.LabelReference("l1")
    subroutine1Ops = [
        pt.TealOp(None, pt.Op.store, 0),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bnz, subroutine1L1Label),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.minus),
        pt.TealOp(None, pt.Op.callsub, subroutine1),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.retsub),
        pt.TealLabel(None, subroutine1L1Label),
        pt.TealOp(None, pt.Op.load, 255),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine2L1Label = pt.LabelReference("l1")
    subroutine2Ops = [
        pt.TealOp(None, pt.Op.store, 1),
        pt.TealOp(None, pt.Op.load, 1),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bnz, subroutine2L1Label),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.minus),
        pt.TealOp(None, pt.Op.callsub, subroutine1),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.retsub),
        pt.TealLabel(None, subroutine2L1Label),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine3Ops = [
        pt.TealOp(None, pt.Op.callsub, subroutine3),
        pt.TealOp(None, pt.Op.retsub),
    ]

    l1Label = pt.LabelReference("l1")
    mainOps = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, 255),
        pt.TealOp(None, pt.Op.txn, "Fee"),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bz, l1Label),
        pt.TealOp(None, pt.Op.int, 100),
        pt.TealOp(None, pt.Op.callsub, subroutine1),
        pt.TealOp(None, pt.Op.return_),
        pt.TealLabel(None, l1Label),
        pt.TealOp(None, pt.Op.int, 101),
        pt.TealOp(None, pt.Op.callsub, subroutine2),
        pt.TealOp(None, pt.Op.return_),
        pt.TealOp(None, pt.Op.callsub, subroutine3),
    ]

    subroutineMapping = {
        None: mainOps,
        subroutine1: subroutine1Ops,
        subroutine2: subroutine2Ops,
        subroutine3: subroutine3Ops,
    }

    subroutineGraph = {
        subroutine1: {subroutine1},
        subroutine2: {subroutine1},
        subroutine3: {subroutine3},
    }

    localSlots = {None: set(), subroutine1: {0}, subroutine2: {1}, subroutine3: {}}

    spillLocalSlotsDuringRecursion(5, subroutineMapping, subroutineGraph, localSlots)

    assert subroutineMapping == {
        None: [
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.store, 255),
            pt.TealOp(None, pt.Op.txn, "Fee"),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bz, l1Label),
            pt.TealOp(None, pt.Op.int, 100),
            pt.TealOp(None, pt.Op.callsub, subroutine1),
            pt.TealOp(None, pt.Op.return_),
            pt.TealLabel(None, l1Label),
            pt.TealOp(None, pt.Op.int, 101),
            pt.TealOp(None, pt.Op.callsub, subroutine2),
            pt.TealOp(None, pt.Op.return_),
            pt.TealOp(None, pt.Op.callsub, subroutine3),
        ],
        subroutine1: [
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bnz, subroutine1L1Label),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.minus),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.swap),
            pt.TealOp(None, pt.Op.callsub, subroutine1),
            pt.TealOp(None, pt.Op.swap),
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.add),
            pt.TealOp(None, pt.Op.retsub),
            pt.TealLabel(None, subroutine1L1Label),
            pt.TealOp(None, pt.Op.load, 255),
            pt.TealOp(None, pt.Op.retsub),
        ],
        subroutine2: [
            pt.TealOp(None, pt.Op.store, 1),
            pt.TealOp(None, pt.Op.load, 1),
            pt.TealOp(None, pt.Op.int, 0),
            pt.TealOp(None, pt.Op.eq),
            pt.TealOp(None, pt.Op.bnz, subroutine2L1Label),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.minus),
            pt.TealOp(None, pt.Op.callsub, subroutine1),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.add),
            pt.TealOp(None, pt.Op.retsub),
            pt.TealLabel(None, subroutine2L1Label),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.retsub),
        ],
        subroutine3: [
            pt.TealOp(None, pt.Op.callsub, subroutine3),
            pt.TealOp(None, pt.Op.retsub),
        ],
    }


def test_spillLocalSlotsDuringRecursion_recursive_many_args_no_return_v4():
    def subImpl(a1, a2, a3):
        return None

    subroutine = pt.SubroutineDefinition(subImpl, pt.TealType.none)

    subroutineOps = [
        pt.TealOp(None, pt.Op.store, 0),
        pt.TealOp(None, pt.Op.store, 1),
        pt.TealOp(None, pt.Op.store, 2),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.retsub),
    ]

    mainOps = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.return_),
    ]

    subroutineMapping = {
        None: mainOps,
        subroutine: subroutineOps,
    }

    subroutineGraph = {
        subroutine: {subroutine},
    }

    localSlots = {None: set(), subroutine: {0, 1, 2}}

    spillLocalSlotsDuringRecursion(4, subroutineMapping, subroutineGraph, localSlots)

    assert subroutineMapping == {
        None: [
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.return_),
        ],
        subroutine: [
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.store, 1),
            pt.TealOp(None, pt.Op.store, 2),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.load, 1),
            pt.TealOp(None, pt.Op.load, 2),
            pt.TealOp(None, pt.Op.dig, 5),
            pt.TealOp(None, pt.Op.dig, 5),
            pt.TealOp(None, pt.Op.dig, 5),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.store, 2),
            pt.TealOp(None, pt.Op.store, 1),
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.pop),
            pt.TealOp(None, pt.Op.pop),
            pt.TealOp(None, pt.Op.pop),
            pt.TealOp(None, pt.Op.retsub),
        ],
    }


def test_spillLocalSlotsDuringRecursion_recursive_many_args_no_return_v5():
    def subImpl(a1, a2, a3):
        return None

    subroutine = pt.SubroutineDefinition(subImpl, pt.TealType.none)

    subroutineOps = [
        pt.TealOp(None, pt.Op.store, 0),
        pt.TealOp(None, pt.Op.store, 1),
        pt.TealOp(None, pt.Op.store, 2),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.retsub),
    ]

    mainOps = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.return_),
    ]

    subroutineMapping = {
        None: mainOps,
        subroutine: subroutineOps,
    }

    subroutineGraph = {
        subroutine: {subroutine},
    }

    localSlots = {None: set(), subroutine: {0, 1, 2}}

    spillLocalSlotsDuringRecursion(5, subroutineMapping, subroutineGraph, localSlots)

    assert subroutineMapping == {
        None: [
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.return_),
        ],
        subroutine: [
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.store, 1),
            pt.TealOp(None, pt.Op.store, 2),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.load, 1),
            pt.TealOp(None, pt.Op.load, 2),
            pt.TealOp(None, pt.Op.uncover, 5),
            pt.TealOp(None, pt.Op.uncover, 5),
            pt.TealOp(None, pt.Op.uncover, 5),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.store, 2),
            pt.TealOp(None, pt.Op.store, 1),
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.retsub),
        ],
    }


def test_spillLocalSlotsDuringRecursion_recursive_many_args_return_v4():
    def subImpl(a1, a2, a3):
        return None

    subroutine = pt.SubroutineDefinition(subImpl, pt.TealType.uint64)

    subroutineOps = [
        pt.TealOp(None, pt.Op.store, 0),
        pt.TealOp(None, pt.Op.store, 1),
        pt.TealOp(None, pt.Op.store, 2),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.retsub),
    ]

    mainOps = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.return_),
    ]

    subroutineMapping = {
        None: mainOps,
        subroutine: subroutineOps,
    }

    subroutineGraph = {
        subroutine: {subroutine},
    }

    localSlots = {None: set(), subroutine: {0, 1, 2}}

    spillLocalSlotsDuringRecursion(4, subroutineMapping, subroutineGraph, localSlots)

    assert subroutineMapping == {
        None: [
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.return_),
        ],
        subroutine: [
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.store, 1),
            pt.TealOp(None, pt.Op.store, 2),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.load, 1),
            pt.TealOp(None, pt.Op.load, 2),
            pt.TealOp(None, pt.Op.dig, 5),
            pt.TealOp(None, pt.Op.dig, 5),
            pt.TealOp(None, pt.Op.dig, 5),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.store, 2),
            pt.TealOp(None, pt.Op.store, 1),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.swap),
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.swap),
            pt.TealOp(None, pt.Op.pop),
            pt.TealOp(None, pt.Op.swap),
            pt.TealOp(None, pt.Op.pop),
            pt.TealOp(None, pt.Op.swap),
            pt.TealOp(None, pt.Op.pop),
            pt.TealOp(None, pt.Op.retsub),
        ],
    }


def test_spillLocalSlotsDuringRecursion_recursive_many_args_return_v5():
    def subImpl(a1, a2, a3):
        return None

    subroutine = pt.SubroutineDefinition(subImpl, pt.TealType.uint64)

    subroutineOps = [
        pt.TealOp(None, pt.Op.store, 0),
        pt.TealOp(None, pt.Op.store, 1),
        pt.TealOp(None, pt.Op.store, 2),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.retsub),
    ]

    mainOps = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.return_),
    ]

    subroutineMapping = {
        None: mainOps,
        subroutine: subroutineOps,
    }

    subroutineGraph = {
        subroutine: {subroutine},
    }

    localSlots = {None: set(), subroutine: {0, 1, 2}}

    spillLocalSlotsDuringRecursion(5, subroutineMapping, subroutineGraph, localSlots)

    assert subroutineMapping == {
        None: [
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.return_),
        ],
        subroutine: [
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.store, 1),
            pt.TealOp(None, pt.Op.store, 2),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.load, 1),
            pt.TealOp(None, pt.Op.load, 2),
            pt.TealOp(None, pt.Op.uncover, 5),
            pt.TealOp(None, pt.Op.uncover, 5),
            pt.TealOp(None, pt.Op.uncover, 5),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.cover, 3),
            pt.TealOp(None, pt.Op.store, 2),
            pt.TealOp(None, pt.Op.store, 1),
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.retsub),
        ],
    }


def test_spillLocalSlotsDuringRecursion_recursive_more_args_than_slots_v5():
    def subImpl(a1, a2, a3):
        return None

    subroutine = pt.SubroutineDefinition(subImpl, pt.TealType.uint64)

    subroutineOps = [
        pt.TealOp(None, pt.Op.store, 0),
        pt.TealOp(None, pt.Op.store, 1),
        pt.TealOp(None, pt.Op.pop),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.retsub),
    ]

    mainOps = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.return_),
    ]

    subroutineMapping = {
        None: mainOps,
        subroutine: subroutineOps,
    }

    subroutineGraph = {
        subroutine: {subroutine},
    }

    localSlots = {None: set(), subroutine: {0, 1}}

    spillLocalSlotsDuringRecursion(5, subroutineMapping, subroutineGraph, localSlots)

    assert subroutineMapping == {
        None: [
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.return_),
        ],
        subroutine: [
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.store, 1),
            pt.TealOp(None, pt.Op.pop),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.cover, 3),
            pt.TealOp(None, pt.Op.load, 1),
            pt.TealOp(None, pt.Op.cover, 3),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.cover, 2),
            pt.TealOp(None, pt.Op.store, 1),
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.retsub),
        ],
    }


def test_spillLocalSlotsDuringRecursion_recursive_more_slots_than_args_v5():
    def subImpl(a1, a2, a3):
        return None

    subroutine = pt.SubroutineDefinition(subImpl, pt.TealType.uint64)

    subroutineOps = [
        pt.TealOp(None, pt.Op.store, 0),
        pt.TealOp(None, pt.Op.store, 1),
        pt.TealOp(None, pt.Op.store, 2),
        pt.TealOp(None, pt.Op.int, 10),
        pt.TealOp(None, pt.Op.store, 3),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.retsub),
    ]

    mainOps = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.int, 2),
        pt.TealOp(None, pt.Op.int, 3),
        pt.TealOp(None, pt.Op.callsub, subroutine),
        pt.TealOp(None, pt.Op.return_),
    ]

    subroutineMapping = {
        None: mainOps,
        subroutine: subroutineOps,
    }

    subroutineGraph = {
        subroutine: {subroutine},
    }

    localSlots = {None: set(), subroutine: {0, 1, 2, 3}}

    spillLocalSlotsDuringRecursion(5, subroutineMapping, subroutineGraph, localSlots)

    assert subroutineMapping == {
        None: [
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.return_),
        ],
        subroutine: [
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.store, 1),
            pt.TealOp(None, pt.Op.store, 2),
            pt.TealOp(None, pt.Op.int, 10),
            pt.TealOp(None, pt.Op.store, 3),
            pt.TealOp(None, pt.Op.int, 1),
            pt.TealOp(None, pt.Op.int, 2),
            pt.TealOp(None, pt.Op.int, 3),
            pt.TealOp(None, pt.Op.load, 0),
            pt.TealOp(None, pt.Op.load, 1),
            pt.TealOp(None, pt.Op.load, 2),
            pt.TealOp(None, pt.Op.load, 3),
            pt.TealOp(None, pt.Op.uncover, 6),
            pt.TealOp(None, pt.Op.uncover, 6),
            pt.TealOp(None, pt.Op.uncover, 6),
            pt.TealOp(None, pt.Op.callsub, subroutine),
            pt.TealOp(None, pt.Op.cover, 4),
            pt.TealOp(None, pt.Op.store, 3),
            pt.TealOp(None, pt.Op.store, 2),
            pt.TealOp(None, pt.Op.store, 1),
            pt.TealOp(None, pt.Op.store, 0),
            pt.TealOp(None, pt.Op.retsub),
        ],
    }


def test_spillLocalSlotsDuringRecursion_recursive_with_scratchvar():
    # modifying test_spillLocalSlotsDuringRecursion_multiple_subroutines_no_recursion()
    # to be recursive and fail due to by-ref args
    def sub1Impl(a1):
        return None

    def sub2Impl(a1, a2: pt.ScratchVar):
        return None

    def sub3Impl(a1, a2, a3):
        return None

    subroutine1 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
    subroutine2 = pt.SubroutineDefinition(sub2Impl, pt.TealType.uint64)
    subroutine3 = pt.SubroutineDefinition(sub3Impl, pt.TealType.none)

    subroutine1L1Label = pt.LabelReference("l1")
    subroutine1Ops = [
        pt.TealOp(None, pt.Op.store, 0),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bnz, subroutine1L1Label),
        pt.TealOp(None, pt.Op.err),
        pt.TealLabel(None, subroutine1L1Label),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.callsub, subroutine3),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine2L1Label = pt.LabelReference("l1")
    subroutine2Ops = [
        pt.TealOp(None, pt.Op.store, 1),
        pt.TealOp(None, pt.Op.load, 1),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bnz, subroutine2L1Label),
        pt.TealOp(None, pt.Op.err),
        pt.TealLabel(None, subroutine2L1Label),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine3Ops = [
        pt.TealOp(None, pt.Op.retsub),
    ]

    l1Label = pt.LabelReference("l1")
    mainOps = [
        pt.TealOp(None, pt.Op.txn, "Fee"),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bz, l1Label),
        pt.TealOp(None, pt.Op.int, 100),
        pt.TealOp(None, pt.Op.callsub, subroutine1),
        pt.TealOp(None, pt.Op.return_),
        pt.TealLabel(None, l1Label),
        pt.TealOp(None, pt.Op.int, 101),
        pt.TealOp(None, pt.Op.callsub, subroutine2),
        pt.TealOp(None, pt.Op.return_),
    ]

    subroutineMapping = {
        None: mainOps,
        subroutine1: subroutine1Ops,
        subroutine2: subroutine2Ops,
        subroutine3: subroutine3Ops,
    }

    subroutineGraph = {
        subroutine1: {subroutine2},
        subroutine2: {subroutine1},
        subroutine3: set(),
    }

    localSlots = {None: set(), subroutine1: {0}, subroutine2: {1}, subroutine3: {}}

    with pytest.raises(pt.TealInputError) as tie:
        spillLocalSlotsDuringRecursion(
            5, subroutineMapping, subroutineGraph, localSlots
        )

    assert (
        "ScratchVar arguments not allowed in recursive subroutines, but a recursive call-path was detected: sub2Impl()-->sub1Impl()-->sub2Impl()"
        in str(tie)
    )


def test_resolveSubroutines():
    def sub1Impl(a1):
        return None

    def sub2Impl(a1, a2):
        return None

    def sub3Impl(a1, a2, a3):
        return None

    subroutine1 = pt.SubroutineDefinition(sub1Impl, pt.TealType.uint64)
    subroutine2 = pt.SubroutineDefinition(sub2Impl, pt.TealType.uint64)
    subroutine3 = pt.SubroutineDefinition(sub3Impl, pt.TealType.none)

    subroutine1L1Label = pt.LabelReference("l1")
    subroutine1Ops = [
        pt.TealOp(None, pt.Op.store, 0),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bnz, subroutine1L1Label),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.minus),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.dig, 1),
        pt.TealOp(None, pt.Op.callsub, subroutine1),
        pt.TealOp(None, pt.Op.swap),
        pt.TealOp(None, pt.Op.store, 0),
        pt.TealOp(None, pt.Op.swap),
        pt.TealOp(None, pt.Op.pop),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.retsub),
        pt.TealLabel(None, subroutine1L1Label),
        pt.TealOp(None, pt.Op.load, 255),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine2L1Label = pt.LabelReference("l1")
    subroutine2Ops = [
        pt.TealOp(None, pt.Op.store, 1),
        pt.TealOp(None, pt.Op.load, 1),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bnz, subroutine2L1Label),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.minus),
        pt.TealOp(None, pt.Op.callsub, subroutine1),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.retsub),
        pt.TealLabel(None, subroutine2L1Label),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    subroutine3Ops = [
        pt.TealOp(None, pt.Op.callsub, subroutine3),
        pt.TealOp(None, pt.Op.retsub),
    ]

    l1Label = pt.LabelReference("l1")
    mainOps = [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, 255),
        pt.TealOp(None, pt.Op.txn, "Fee"),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bz, l1Label),
        pt.TealOp(None, pt.Op.int, 100),
        pt.TealOp(None, pt.Op.callsub, subroutine1),
        pt.TealOp(None, pt.Op.return_),
        pt.TealLabel(None, l1Label),
        pt.TealOp(None, pt.Op.int, 101),
        pt.TealOp(None, pt.Op.callsub, subroutine2),
        pt.TealOp(None, pt.Op.return_),
        pt.TealOp(None, pt.Op.callsub, subroutine3),
    ]

    subroutineMapping = {
        None: mainOps,
        subroutine1: subroutine1Ops,
        subroutine2: subroutine2Ops,
        subroutine3: subroutine3Ops,
    }

    expected = OrderedDict()
    expected[subroutine1] = "sub1Impl_0"
    expected[subroutine2] = "sub2Impl_1"
    expected[subroutine3] = "sub3Impl_2"

    actual = resolveSubroutines(subroutineMapping)
    assert actual == expected

    assert mainOps == [
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.store, 255),
        pt.TealOp(None, pt.Op.txn, "Fee"),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bz, l1Label),
        pt.TealOp(None, pt.Op.int, 100),
        pt.TealOp(None, pt.Op.callsub, expected[subroutine1]),
        pt.TealOp(None, pt.Op.return_),
        pt.TealLabel(None, l1Label),
        pt.TealOp(None, pt.Op.int, 101),
        pt.TealOp(None, pt.Op.callsub, expected[subroutine2]),
        pt.TealOp(None, pt.Op.return_),
        pt.TealOp(None, pt.Op.callsub, expected[subroutine3]),
    ]

    assert subroutine1Ops == [
        pt.TealOp(None, pt.Op.store, 0),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bnz, subroutine1L1Label),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.minus),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.dig, 1),
        pt.TealOp(None, pt.Op.callsub, expected[subroutine1]),
        pt.TealOp(None, pt.Op.swap),
        pt.TealOp(None, pt.Op.store, 0),
        pt.TealOp(None, pt.Op.swap),
        pt.TealOp(None, pt.Op.pop),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.retsub),
        pt.TealLabel(None, subroutine1L1Label),
        pt.TealOp(None, pt.Op.load, 255),
        pt.TealOp(None, pt.Op.retsub),
    ]

    assert subroutine2Ops == [
        pt.TealOp(None, pt.Op.store, 1),
        pt.TealOp(None, pt.Op.load, 1),
        pt.TealOp(None, pt.Op.int, 0),
        pt.TealOp(None, pt.Op.eq),
        pt.TealOp(None, pt.Op.bnz, subroutine2L1Label),
        pt.TealOp(None, pt.Op.load, 0),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.minus),
        pt.TealOp(None, pt.Op.callsub, expected[subroutine1]),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.add),
        pt.TealOp(None, pt.Op.retsub),
        pt.TealLabel(None, subroutine2L1Label),
        pt.TealOp(None, pt.Op.int, 1),
        pt.TealOp(None, pt.Op.retsub),
    ]

    assert subroutine3Ops == [
        pt.TealOp(None, pt.Op.callsub, expected[subroutine3]),
        pt.TealOp(None, pt.Op.retsub),
    ]



from typing import (
    Union,
    Sequence,
    TypeVar,
    Generic,
    Final,
    cast,
)
from abc import abstractmethod

from pyteal.types import TealType, require_type
from pyteal.errors import TealInputError
from pyteal.ast.expr import Expr
from pyteal.ast.seq import Seq
from pyteal.ast.int import Int
from pyteal.ast.if_ import If
from pyteal.ast.unaryexpr import Len
from pyteal.ast.binaryexpr import ExtractUint16
from pyteal.ast.naryexpr import Concat

from pyteal.ast.abi.type import TypeSpec, BaseType, ComputedValue
from pyteal.ast.abi.tuple import _encode_tuple
from pyteal.ast.abi.bool import Bool, BoolTypeSpec
from pyteal.ast.abi.uint import Uint16, Uint16TypeSpec
from pyteal.ast.abi.util import substring_for_decoding

T = TypeVar("T", bound=BaseType)


class ArrayTypeSpec(TypeSpec, Generic[T]):
    """The abstract base class for both static and dynamic array TypeSpecs."""

    def __init__(self, value_type_spec: TypeSpec) -> None:
        super().__init__()
        self.value_spec: Final = value_type_spec

    def value_type_spec(self) -> TypeSpec:
        """Get the TypeSpec of the value type this array can hold."""
        return self.value_spec

    def storage_type(self) -> TealType:
        return TealType.bytes

    @abstractmethod
    def is_length_dynamic(self) -> bool:
        """Check if this array has a dynamic or static length."""
        pass

    def _stride(self) -> int:
        """Get the "stride" of this array.

        The stride is defined as the byte length of each element in the array's encoded "head"
        portion.

        If the underlying value type is static, then the stride is the static byte length of that
        type. Otherwise, the stride is the static byte length of a Uint16 (2 bytes).
        """
        if self.value_spec.is_dynamic():
            return Uint16TypeSpec().byte_length_static()
        return self.value_spec.byte_length_static()


ArrayTypeSpec.__module__ = "pyteal.abi"


class Array(BaseType, Generic[T]):
    """The abstract base class for both ABI static and dynamic array instances.

    This class contains basic implementations of ABI array methods, including:
      * basic array elements setting method
      * basic encoding and decoding of ABI array
      * item retrieving by index (expression or integer)
    """

    def __init__(self, spec: ArrayTypeSpec) -> None:
        super().__init__(spec)

    def type_spec(self) -> ArrayTypeSpec[T]:
        return cast(ArrayTypeSpec, super().type_spec())

    def decode(
        self,
        encoded: Expr,
        *,
        start_index: Expr = None,
        end_index: Expr = None,
        length: Expr = None
    ) -> Expr:
        """Decode a substring of the passed in encoded byte string and set it as this type's value.

        Args:
            encoded: An expression containing the bytes to decode. Must evaluate to TealType.bytes.
            start_index (optional): An expression containing the index to start decoding. Must
                evaluate to TealType.uint64. Defaults to None.
            end_index (optional): An expression containing the index to stop decoding. Must evaluate
                to TealType.uint64. Defaults to None.
            length (optional): An expression containing the length of the substring to decode. Must
                evaluate to TealType.uint64. Defaults to None.

        Returns:
            An expression that partitions the needed parts from given byte strings and stores into
            the scratch variable.
        """
        extracted = substring_for_decoding(
            encoded, start_index=start_index, end_index=end_index, length=length
        )
        return self.stored_value.store(extracted)

    def set(self, values: Sequence[T]) -> Expr:
        """Set the ABI array with a sequence of ABI type variables.

        The function first type-check the argument `values` to make sure the sequence of ABI type
        variables before storing them to the underlying ScratchVar. If any of the input element does
        not match expected array element type, error would be raised about type-mismatch.

        If static length of array is not available, this function would
        * infer the array length from the sequence element number.
        * store the inferred array length in uint16 format.
        * concatenate the encoded array length at the beginning of array encoding.

        Args:
            values: The sequence of ABI type variables to store in ABI array.

        Returns:
            A PyTeal expression that stores encoded sequence of ABI values in its internal
            ScratchVar.
        """
        for index, value in enumerate(values):
            if self.type_spec().value_type_spec() != value.type_spec():
                raise TealInputError(
                    "Cannot assign type {} at index {} to {}".format(
                        value.type_spec(),
                        index,
                        self.type_spec().value_type_spec(),
                    )
                )

        encoded = _encode_tuple(values)

        if self.type_spec().is_length_dynamic():
            length_tmp = Uint16()
            length_prefix = Seq(length_tmp.set(len(values)), length_tmp.encode())
            encoded = Concat(length_prefix, encoded)

        return self.stored_value.store(encoded)

    def encode(self) -> Expr:
        """Encode the ABI array to be a byte string.

        Returns:
            A PyTeal expression that encodes this ABI array to a byte string.
        """
        return self.stored_value.load()

    @abstractmethod
    def length(self) -> Expr:
        """Get the element number of this ABI array.

        Returns:
            A PyTeal expression that represents the array length.
        """
        pass

    def __getitem__(self, index: Union[int, Expr]) -> "ArrayElement[T]":
        """Retrieve an element by its index in this array.

        Indexes start at 0.

        Args:
            index: either a Python integer or a PyTeal expression that evaluates to a TealType.uint64.
                If a Python integer is used, this function will raise an error if its value is negative.
                In either case, if the index is outside of the bounds of this array, the program will
                fail at runtime.

        Returns:
            An ArrayElement that corresponds to the element at the given index. This type is a ComputedValue.
        """
        if type(index) is int:
            if index < 0:
                raise TealInputError("Index out of bounds: {}".format(index))
            index = Int(index)
        return ArrayElement(self, cast(Expr, index))


Array.__module__ = "pyteal.abi"


class ArrayElement(ComputedValue[T]):
    """The class that represents an ABI array element.

    This class requires a reference to the array that the array element belongs to, and a PyTeal
    expression (required to be TealType.uint64) which contains the array index.
    """

    def __init__(self, array: Array[T], index: Expr) -> None:
        """Creates a new ArrayElement.

        Args:
            array: The ABI array that the array element belongs to.
            index: A PyTeal expression (required to be TealType.uint64) stands for array index.
        """
        super().__init__()
        require_type(index, TealType.uint64)
        self.array = array
        self.index = index

    def produced_type_spec(self) -> TypeSpec:
        return self.array.type_spec().value_type_spec()

    def store_into(self, output: T) -> Expr:
        """Partitions the byte string of the given ABI array and stores the byte string of array
        element in the ABI value output.

        The function first checks if the output type matches with array element type, and throw
        error if type-mismatch.

        Args:
            output: An ABI typed value that the array element byte string stores into.

        Returns:
            An expression that stores the byte string of the array element into value `output`.
        """
        if output.type_spec() != self.produced_type_spec():
            raise TealInputError("Output type does not match value type")

        encodedArray = self.array.encode()
        arrayType = self.array.type_spec()

        # If the array element type is Bool, we compute the bit index
        # (if array is dynamic we add 16 to bit index for dynamic array length uint16 prefix)
        # and decode bit with given array encoding and the bit index for boolean bit.
        if output.type_spec() == BoolTypeSpec():
            bitIndex = self.index
            if arrayType.is_dynamic():
                bitIndex = bitIndex + Int(Uint16TypeSpec().bit_size())
            return cast(Bool, output).decode_bit(encodedArray, bitIndex)

        # Compute the byteIndex (first byte indicating the element encoding)
        # (If the array is dynamic, add 2 to byte index for dynamic array length uint16 prefix)
        byteIndex = Int(arrayType._stride()) * self.index
        if arrayType.is_length_dynamic():
            byteIndex = byteIndex + Int(Uint16TypeSpec().byte_length_static())

        arrayLength = self.array.length()

        # Handling case for array elements are dynamic:
        # * `byteIndex` is pointing at the uint16 byte encoding indicating the beginning offset of
        #   the array element byte encoding.
        #
        # * `valueStart` is extracted from the uint16 bytes pointed by `byteIndex`.
        #
        # * If `index == arrayLength - 1` (last element in array), `valueEnd` is pointing at the
        #   end of the array byte encoding.
        #
        # * otherwise, `valueEnd` is inferred from `nextValueStart`, which is the beginning offset
        #   of the next array element byte encoding.
        if arrayType.value_type_spec().is_dynamic():
            valueStart = ExtractUint16(encodedArray, byteIndex)
            nextValueStart = ExtractUint16(
                encodedArray, byteIndex + Int(Uint16TypeSpec().byte_length_static())
            )
            if arrayType.is_length_dynamic():
                valueStart = valueStart + Int(Uint16TypeSpec().byte_length_static())
                nextValueStart = nextValueStart + Int(
                    Uint16TypeSpec().byte_length_static()
                )

            valueEnd = (
                If(self.index + Int(1) == arrayLength)
                .Then(Len(encodedArray))
                .Else(nextValueStart)
            )

            return output.decode(
                encodedArray, start_index=valueStart, end_index=valueEnd
            )

        # Handling case for array elements are static:
        # since array._stride() is element's static byte length
        # we partition the substring for array element.
        valueStart = byteIndex
        valueLength = Int(arrayType._stride())
        return output.decode(encodedArray, start_index=valueStart, length=valueLength)


ArrayElement.__module__ = "pyteal.abi"



from typing import List, Union

from pyteal.ast.multi import MultiValue

from pyteal.types import TealType
from pyteal.ir import Op
from pyteal.ast.expr import Expr
from pyteal.ast.scratch import ScratchLoad, ScratchSlot


class MaybeValue(MultiValue):
    """Represents a get operation returning a value that may not exist."""

    def __init__(
        self,
        op: Op,
        type: TealType,
        *,
        immediate_args: List[Union[int, str]] = None,
        args: List[Expr] = None
    ):
        """Create a new MaybeValue.

        Args:
            op: The operation that returns values.
            type: The type of the returned value.
            immediate_args (optional): Immediate arguments for the op. Defaults to None.
            args (optional): Stack arguments for the op. Defaults to None.
        """
        types = [type, TealType.uint64]
        super().__init__(op, types, immediate_args=immediate_args, args=args)

    def hasValue(self) -> ScratchLoad:
        """Check if the value exists.

        This will return 1 if the value exists, otherwise 0.
        """
        return self.output_slots[1].load(self.types[1])

    def value(self) -> ScratchLoad:
        """Get the value.

        If the value exists, it will be returned. Otherwise, the zero value for this type will be
        returned (i.e. either 0 or an empty byte string, depending on the type).
        """
        return self.output_slots[0].load(self.types[0])

    @property
    def slotOk(self) -> ScratchSlot:
        """Get the scratch slot that stores hasValue.

        Note: This is mainly added for backwards compatability and normally shouldn't be used
        directly in pyteal code.
        """
        return self.output_slots[1]

    @property
    def slotValue(self) -> ScratchSlot:
        """Get the scratch slot that stores the value or the zero value for the type if the value
        doesn't exist.

        Note: This is mainly added for backwards compatability and normally shouldn't be used
        directly in pyteal code.
        """
        return self.output_slots[0]


MaybeValue.__module__ = "pyteal"



from typing import TYPE_CHECKING

from pyteal.errors import (
    TealCompileError,
    TealInputError,
)
from pyteal.types import TealType, require_type
from pyteal.ir import TealSimpleBlock, TealConditionalBlock
from pyteal.ast.expr import Expr
from pyteal.ast.seq import _use_seq_if_multiple

if TYPE_CHECKING:
    from pyteal.compiler import CompileOptions


class If(Expr):
    """Simple two-way conditional expression."""

    def __init__(
        self, cond: Expr, thenBranch: Expr = None, elseBranch: Expr = None
    ) -> None:
        """Create a new If expression.

        When this If expression is executed, the condition will be evaluated, and if it produces a
        true value, thenBranch will be executed and used as the return value for this expression.
        Otherwise, elseBranch will be executed and used as the return value, if it is provided.

        Args:
            cond: The condition to check. Must evaluate to uint64.
            thenBranch: Expression to evaluate if the condition is true.
            elseBranch (optional): Expression to evaluate if the condition is false. Must evaluate
                to the same type as thenBranch, if provided. Defaults to None.
        """
        super().__init__()
        require_type(cond, TealType.uint64)
        # Flag to denote and check whether the new If().Then() syntax is being used
        self.alternateSyntaxFlag = False

        if thenBranch:
            if elseBranch:
                require_type(thenBranch, elseBranch.type_of())
            else:
                # If there is only a thenBranch, then it should evaluate to none type
                require_type(thenBranch, TealType.none)
        else:
            self.alternateSyntaxFlag = True

        self.cond = cond
        self.thenBranch = thenBranch
        self.elseBranch = elseBranch

    def __teal__(self, options: "CompileOptions"):
        if self.thenBranch is None:
            raise TealCompileError("If expression must have a thenBranch", self)

        condStart, condEnd = self.cond.__teal__(options)
        thenStart, thenEnd = self.thenBranch.__teal__(options)
        end = TealSimpleBlock([])

        branchBlock = TealConditionalBlock([])
        branchBlock.setTrueBlock(thenStart)

        condEnd.setNextBlock(branchBlock)
        thenEnd.setNextBlock(end)

        if self.elseBranch is None:
            branchBlock.setFalseBlock(end)
        else:
            elseStart, elseEnd = self.elseBranch.__teal__(options)
            branchBlock.setFalseBlock(elseStart)
            elseEnd.setNextBlock(end)

        return condStart, end

    def __str__(self):
        if self.thenBranch is None:
            raise TealCompileError("If expression must have a thenBranch", self)
        if self.elseBranch is None:
            return "(If {} {})".format(self.cond, self.thenBranch)
        return "(If {} {} {})".format(self.cond, self.thenBranch, self.elseBranch)

    def type_of(self):
        if self.thenBranch is None:
            raise TealCompileError("If expression must have a thenBranch", self)

        if self.elseBranch is None:
            # if there is only a thenBranch, it must evaluate to TealType.none
            require_type(self.thenBranch, TealType.none)

        return self.thenBranch.type_of()

    def has_return(self):
        if self.thenBranch is None:
            raise TealCompileError("If expression must have a thenBranch", self)

        if self.elseBranch is None:
            # return false in this case because elseBranch does not exist, so it can't have a return
            # op
            return False
        # otherwise, this expression has a return op only if both branches result in a return op
        return self.thenBranch.has_return() and self.elseBranch.has_return()

    def Then(self, thenBranch: Expr, *then_branch_multi: Expr):
        if not self.alternateSyntaxFlag:
            raise TealInputError("Cannot mix two different If syntax styles")

        thenBranch = _use_seq_if_multiple(thenBranch, *then_branch_multi)

        if not self.elseBranch:
            self.thenBranch = thenBranch
        else:
            if not isinstance(self.elseBranch, If):
                raise TealInputError("Else-Then block is malformed")
            self.elseBranch.Then(thenBranch)
        return self

    def ElseIf(self, cond):
        if not self.alternateSyntaxFlag:
            raise TealInputError("Cannot mix two different If syntax styles")

        if not self.elseBranch:
            self.elseBranch = If(cond)
        else:
            if not isinstance(self.elseBranch, If):
                raise TealInputError("Else-ElseIf block is malformed")
            self.elseBranch.ElseIf(cond)
        return self

    def Else(self, elseBranch: Expr, *else_branch_multi: Expr):
        if not self.alternateSyntaxFlag:
            raise TealInputError("Cannot mix two different If syntax styles")

        if not self.thenBranch:
            raise TealInputError("Must set Then branch before Else branch")

        elseBranch = _use_seq_if_multiple(elseBranch, *else_branch_multi)

        if not self.elseBranch:
            require_type(elseBranch, self.thenBranch.type_of())
            self.elseBranch = elseBranch
        else:
            if not isinstance(self.elseBranch, If):
                raise TealInputError("Else-Else block is malformed")
            self.elseBranch.Else(elseBranch)
        return self


If.__module__ = "pyteal"



from itertools import product
from pathlib import Path
import pytest
from typing import Literal, Optional, Tuple

import pyteal as pt

from tests.blackbox import Blackbox, BlackboxWrapper, PyTealDryRunExecutor

from tests.compile_asserts import assert_teal_as_expected

PATH = Path.cwd() / "tests" / "unit"
FIXTURES = PATH / "teal"
GENERATED = PATH / "generated"

# ---- Subroutine Unit Test Examples ---- #


@Blackbox(input_types=[])
@pt.Subroutine(pt.TealType.none)
def utest_noop():
    return pt.Pop(pt.Int(0))


@Blackbox(input_types=[pt.TealType.uint64, pt.TealType.bytes, pt.TealType.anytype])
@pt.Subroutine(pt.TealType.none)
def utest_noop_args(x, y, z):
    return pt.Pop(pt.Int(0))


@Blackbox(input_types=[])
@pt.Subroutine(pt.TealType.uint64)
def utest_int():
    return pt.Int(0)


@Blackbox(input_types=[pt.TealType.uint64, pt.TealType.bytes, pt.TealType.anytype])
@pt.Subroutine(pt.TealType.uint64)
def utest_int_args(x, y, z):
    return pt.Int(0)


@Blackbox(input_types=[])
@pt.Subroutine(pt.TealType.bytes)
def utest_bytes():
    return pt.Bytes("")


@Blackbox(input_types=[pt.TealType.uint64, pt.TealType.bytes, pt.TealType.anytype])
@pt.Subroutine(pt.TealType.bytes)
def utest_bytes_args(x, y, z):
    return pt.Bytes("")


@Blackbox(input_types=[])
@pt.Subroutine(pt.TealType.anytype)
def utest_any():
    x = pt.ScratchVar(pt.TealType.anytype)
    return pt.Seq(x.store(pt.Int(0)), x.load())


@Blackbox(input_types=[pt.TealType.uint64, pt.TealType.bytes, pt.TealType.anytype])
@pt.Subroutine(pt.TealType.anytype)
def utest_any_args(x, y, z):
    x = pt.ScratchVar(pt.TealType.anytype)
    return pt.Seq(x.store(pt.Int(0)), x.load())


UNITS = [
    utest_noop,
    utest_noop_args,
    utest_int,
    utest_int_args,
    utest_bytes,
    utest_bytes_args,
    utest_any,
    utest_any_args,
]


# ---- ABI Return Subroutine Unit Test Examples ---- #


@Blackbox(input_types=[])
@pt.ABIReturnSubroutine
def fn_0arg_0ret() -> pt.Expr:
    return pt.Return()


@Blackbox(input_types=[])
@pt.ABIReturnSubroutine
def fn_0arg_uint64_ret(*, output: pt.abi.Uint64) -> pt.Expr:
    return output.set(1)


@Blackbox(input_types=[None])
@pt.ABIReturnSubroutine
def fn_1arg_0ret(a: pt.abi.Uint64) -> pt.Expr:
    return pt.Return()


@Blackbox(input_types=[None])
@pt.ABIReturnSubroutine
def fn_1arg_1ret(a: pt.abi.Uint64, *, output: pt.abi.Uint64) -> pt.Expr:
    return output.set(a)


@Blackbox(input_types=[None, None])
@pt.ABIReturnSubroutine
def fn_2arg_0ret(
    a: pt.abi.Uint64, b: pt.abi.StaticArray[pt.abi.Byte, Literal[10]]
) -> pt.Expr:
    return pt.Return()


@Blackbox(input_types=[pt.TealType.bytes])
@pt.ABIReturnSubroutine
def fn_1tt_arg_uint64_ret(x, *, output: pt.abi.Uint64) -> pt.Expr:
    return output.set(1)


@Blackbox(input_types=[None, pt.TealType.uint64, None])
@pt.ABIReturnSubroutine
def fn_3mixed_args_0ret(
    a: pt.abi.Uint64, b: pt.ScratchVar, C: pt.abi.StaticArray[pt.abi.Byte, Literal[10]]
) -> pt.Expr:
    return pt.Return()


@Blackbox(input_types=[None, pt.TealType.bytes])
@pt.ABIReturnSubroutine
def fn_2mixed_arg_1ret(
    a: pt.abi.Uint64, b: pt.ScratchVar, *, output: pt.abi.Uint64
) -> pt.Expr:
    return pt.Seq(b.store(a.encode()), output.set(a))


ABI_UNITS = [
    (fn_0arg_0ret, None),
    (fn_0arg_uint64_ret, pt.abi.Uint64()),
    (fn_1arg_0ret, None),
    (fn_1arg_1ret, pt.abi.Uint64()),
    (fn_2arg_0ret, None),
    (fn_1tt_arg_uint64_ret, pt.abi.Uint64()),
    (fn_3mixed_args_0ret, None),
    (fn_2mixed_arg_1ret, pt.abi.Uint64()),
]


# ---- test functions ---- #


@pytest.mark.parametrize("subr, mode", product(UNITS, pt.Mode))
def test_blackbox_pyteal(subr: BlackboxWrapper, mode: pt.Mode):
    is_app = mode == pt.Mode.Application
    name = f"{'app' if is_app else 'lsig'}_{subr.name()}"

    compiled = PyTealDryRunExecutor(subr, mode).compile(version=6)
    tealdir = GENERATED / "blackbox"
    tealdir.mkdir(parents=True, exist_ok=True)
    save_to = tealdir / (name + ".teal")
    with open(save_to, "w") as f:
        f.write(compiled)

    assert_teal_as_expected(save_to, FIXTURES / "blackbox" / (name + ".teal"))


@pytest.mark.parametrize("subr_abi, mode", product(ABI_UNITS, pt.Mode))
def test_abi_blackbox_pyteal(
    subr_abi: Tuple[BlackboxWrapper, Optional[pt.ast.abi.BaseType]], mode: pt.Mode
):
    subr, abi_return_type = subr_abi
    name = f"{'app' if mode == pt.Mode.Application else 'lsig'}_{subr.name()}"
    print(f"Case {subr.name()=}, {abi_return_type=}, {mode=} ------> {name=}")

    pdre = PyTealDryRunExecutor(subr, mode)
    assert pdre.is_abi(), "should be an ABI subroutine"

    arg_types = pdre.abi_argument_types()
    if subr.name() != "fn_1tt_arg_uint64_ret":
        assert not arg_types or any(
            arg_types
        ), "abi_argument_types() should have had some abi info"

    if abi_return_type:
        expected_sdk_return_type = pt.abi.algosdk_from_type_spec(
            abi_return_type.type_spec()
        )
        assert expected_sdk_return_type == pdre.abi_return_type()
    else:
        assert pdre.abi_return_type() is None

    compiled = pdre.compile(version=6)
    tealdir = GENERATED / "abi"
    tealdir.mkdir(parents=True, exist_ok=True)
    save_to = tealdir / (name + ".teal")
    with open(save_to, "w") as f:
        f.write(compiled)

    assert_teal_as_expected(save_to, FIXTURES / "abi" / (name + ".teal"))


@pytest.mark.parametrize("mode", (pt.Mode.Application, pt.Mode.Signature))
@pytest.mark.parametrize(
    "fn, expected_is_abi", ((utest_noop, False), (fn_0arg_uint64_ret, True))
)
def test_PyTealBlackboxExecutor_is_abi(
    mode: pt.Mode, fn: BlackboxWrapper, expected_is_abi: bool
):
    p = PyTealDryRunExecutor(fn, mode)
    assert p.is_abi() == expected_is_abi
    if expected_is_abi:
        assert p.abi_argument_types() is not None
        assert p.abi_return_type() is not None
    else:
        assert p.abi_argument_types() is None
        assert p.abi_return_type() is None


@pytest.mark.parametrize("mode", (pt.Mode.Application, pt.Mode.Signature))
@pytest.mark.parametrize(
    "fn, expected_arg_count",
    (
        (fn_0arg_uint64_ret, 0),
        (fn_1arg_0ret, 1),
        (fn_1arg_1ret, 1),
        (fn_2arg_0ret, 2),
        (fn_2mixed_arg_1ret, 2),
    ),
)
def test_PyTealBlackboxExecutor_abi_argument_types(
    mode: pt.Mode, fn: BlackboxWrapper, expected_arg_count: int
):
    actual = PyTealDryRunExecutor(fn, mode).abi_argument_types()
    assert actual is not None
    assert len(actual) == expected_arg_count


@pytest.mark.parametrize("mode", (pt.Mode.Application, pt.Mode.Signature))
@pytest.mark.parametrize(
    "fn, expected_does_produce_type",
    (
        (fn_0arg_uint64_ret, True),
        (fn_1arg_0ret, False),
        (fn_1arg_1ret, True),
        (fn_2arg_0ret, False),
        (fn_2mixed_arg_1ret, True),
    ),
)
def test_PyTealBlackboxExecutor_abi_return_type(
    mode: pt.Mode, fn: BlackboxWrapper, expected_does_produce_type: bool
):
    if expected_does_produce_type:
        assert PyTealDryRunExecutor(fn, mode).abi_return_type() is not None
    else:
        assert PyTealDryRunExecutor(fn, mode).abi_return_type() is None
